"""Module with views.

Such as:
HyperEdgeView, CellView, SimplexView, NodeView.
"""

from abc import ABC, abstractmethod
from collections.abc import Collection, Hashable, Iterable, Iterator, Sequence
from itertools import chain
from typing import Any, Literal

from toponetx.classes.cell import Cell
from toponetx.classes.complex import Atom
from toponetx.classes.hyperedge import HyperEdge
from toponetx.classes.path import Path
from toponetx.classes.simplex import Simplex
from toponetx.classes.simplex_trie import SimplexTrie

__all__ = [
    "AtomView",
    "CellView",
    "ColoredHyperEdgeView",
    "HyperEdgeView",
    "NodeView",
    "PathView",
    "SimplexView",
]


class AtomView[T_Atom: Atom](ABC):
    """Abstract class representing a read-only view on a collection of atoms."""

    @abstractmethod
    def __contains__(self, atom: Any) -> bool:
        """Check if a given element is in the view.

        Parameters
        ----------
        atom : Any
            The element to check.

        Returns
        -------
        bool
            Whether the element is in the view.
        """

    @abstractmethod
    def __getitem__(self, atom: Any) -> dict:
        """Get the attributes of a given element.

        Parameters
        ----------
        atom : Any
            The element of interest.

        Returns
        -------
        dict
            The attributes associated with the element.

        Raises
        ------
        KeyError
            If the element is not in the view.
        """

    @abstractmethod
    def __iter__(self) -> Iterator[T_Atom]:
        """Iterate over all elements in the view.

        Returns
        -------
        Iterator
            Iterator to iterate over all elements in the view.
        """

    @abstractmethod
    def __len__(self) -> int:
        """Return the number of elements in the view.

        Returns
        -------
        int
            The number of elements in the view.
        """


class CellView(AtomView[Cell]):
    """A CellView class for cells of a CellComplex."""

    # Dictionary to hold cells, with keys being the tuple that defines the cell, and
    # values being the cell objects
    _cells: dict[tuple[Hashable, ...], Cell]

    def __init__(self) -> None:
        self._cells = {}

    def __getitem__(self, cell: Any) -> dict[Hashable, Any]:
        """Return the attributes of a given cell.

        Parameters
        ----------
        cell : Any
            The cell of interest.

        Returns
        -------
        dict[Hashable, Any]
            The attributes associated with the cell.

        Raises
        ------
        KeyError
            If the cell is not in the cell dictionary.
        """
        if isinstance(cell, Cell):
            cell = tuple(cell.elements)
        elif isinstance(cell, Iterable):
            cell = tuple(cell)
        else:
            raise TypeError("Input must be a tuple, list or a cell.")

        if cell in self._cells:
            return self._cells[cell]._attributes

        raise KeyError(f"cell {cell} is not in the cell dictionary")

    def raw(self, cell: Sequence) -> Cell | list[Cell]:
        """Index the raw cell objects analogous to the overall index of CellView.

        Parameters
        ----------
        cell : Sequence
            The cell of interest.

        Returns
        -------
        Cell or list of Cells
            The raw Cell objects.
            If more than one cell with the same boundary exists, returns a list;
            otherwise a single cell.

        Raises
        ------
        KeyError
            If the cell is not in the cell dictionary.
        """
        if isinstance(cell, Cell):
            cell = cell.elements
        elif isinstance(cell, Sequence) and not isinstance(cell, str | bytes):
            cell = tuple(cell)
        else:
            raise TypeError("Input must be a tuple, list or a cell object.")

        if cell in self._cells:
            return self._cells[cell]
        raise KeyError(f"cell {cell} is not in the cell dictionary")

    def __len__(self) -> int:
        """Return the number of cells in the cell view.

        Returns
        -------
        int
            The number of cells in the cell view.
        """
        return len(self._cells)

    def __iter__(self) -> Iterator[Cell]:
        """Iterate over all cells in the cell view.

        Returns
        -------
        Iterator
            Iterator to iterate over all cells in the cell view.
        """
        return iter(self._cells.values())

    def __contains__(self, atom: Any) -> bool:
        """Check if a given element is in the cell view.

        Parameters
        ----------
        atom : Any
            The element to check.

        Returns
        -------
        bool
            Whether the element is in the cell view.
        """
        if not isinstance(atom, Cell | tuple | list):
            return False

        atom = Cell(atom)
        return any(atom.is_homotopic_to(x) for x in self._cells)

    def __repr__(self) -> str:
        """Return a string representation of the cell view.

        Returns
        -------
        str
            The __repr__ representation of the cell view.
        """
        return f"CellView({[self._cells[cell] for cell in self._cells]})"

    def __str__(self) -> str:
        """Return a string representation of the cell view.

        Returns
        -------
        str
            The __str__ representation of the cell view.
        """
        return f"CellView({[self._cells[cell] for cell in self._cells]})"


class ColoredHyperEdgeView(AtomView):
    """A class for viewing the cells/hyperedges of a colored hypergraph.

    Provides methods for accessing, and retrieving
    information about the cells/hyperedges of a complex.

    Examples
    --------
    >>> hev = tnx.ColoredHyperEdgeView()
    """

    def __init__(self) -> None:
        self.hyperedge_dict = {}

    def __getitem__(self, atom: Any) -> dict[Hashable, Any]:
        """Return the user-defined attributes associated with the given hyperedge.

        Parameters
        ----------
        atom : Any
            The hyperedge for which to return the associated user-defined attributes.

        Returns
        -------
        dict[Hashable, Any]
            The user-defined attributes associated with the given atom.

        Raises
        ------
        KeyError
            If the hyperedge does not exist.
        """
        if isinstance(atom, Hashable) and not isinstance(atom, Collection):
            atom = (atom,)

        if len(atom) == 0:
            raise KeyError(f"Hyperedge {atom} is not in the complex.")
        if len(atom) == 2:
            if isinstance(atom, HyperEdge):
                hyperedge_elements = atom.elements
                key = 0
            elif isinstance(atom[0], Iterable) and isinstance(atom[1], int):
                hyperedge_elements_ = atom[0]
                if not isinstance(hyperedge_elements_, HyperEdge):
                    hyperedge_elements, key = atom
                else:
                    _, key = atom
                    hyperedge_elements = hyperedge_elements_.elements
            else:
                hyperedge_elements = atom
                key = 0
        else:
            hyperedge_elements = atom
            key = 0

        if not isinstance(hyperedge_elements, Iterable) or len(hyperedge_elements) == 0:
            raise KeyError(f"Hyperedge {atom} is not in the complex.")

        for i in self.allranks:
            if frozenset(hyperedge_elements) in self.hyperedge_dict[i]:
                return self.hyperedge_dict[i][frozenset(hyperedge_elements)][key]
        raise KeyError(f"Hyperedge {atom} is not in the complex.")

    @property
    def shape(self) -> tuple[int, ...]:
        """Compute shape.

        Returns
        -------
        tuple[int, ...]
            The shape of the ColoredHyperEdge.
        """
        shape = []
        for i in self.allranks:
            sm = sum(len(self.hyperedge_dict[i][k]) for k in self.hyperedge_dict[i])
            shape.append(sm)
        return tuple(shape)

    def __len__(self) -> int:
        """Compute the number of nodes.

        Returns
        -------
        int
            The number of nodes in the ColoredHyperEdge.
        """
        return sum(self.shape[1:])

    def __iter__(self) -> Iterator:
        """Iterate over the hyperedges.

        Returns
        -------
        Iterator
            The iterator to iterate over the hyperedges.
        """
        lst = []
        for r in self.hyperedge_dict:
            if r == 0:
                continue

            lst.extend(
                (he, k)
                for he in self.hyperedge_dict[r]
                for k in self.hyperedge_dict[r][he]
            )
        return iter(lst)

    def __contains__(self, atom: Any) -> bool:
        """Check if hyperedge is in the hyperedges.

        Parameters
        ----------
        atom : Any
            The hyperedge to check.

        Returns
        -------
        bool
            Return `True` if the hyperedge is contained within the hyperedges.

        Notes
        -----
        Assumption of input here hyperedge = ( elements of hyperedge, key of hyperedge)
        """
        if isinstance(atom, Hashable) and not isinstance(atom, Collection):
            atom = (atom,)

        if len(atom) == 0:
            return False
        if len(atom) == 2:
            if isinstance(atom, HyperEdge):
                hyperedge_elements = atom.elements
                key = 0
            elif isinstance(atom[0], Iterable) and isinstance(atom[1], int):
                hyperedge_elements_ = atom[0]
                if not isinstance(hyperedge_elements_, HyperEdge):
                    hyperedge_elements, key = atom
                else:
                    _, key = atom
                    hyperedge_elements = hyperedge_elements_.elements
            else:
                hyperedge_elements = atom
                key = 0
        else:
            hyperedge_elements = atom
            key = 0

        if not isinstance(hyperedge_elements, Iterable) or len(hyperedge_elements) == 0:
            return False

        for i in self.allranks:
            if frozenset(hyperedge_elements) in self.hyperedge_dict[i]:
                return key in self.hyperedge_dict[i][frozenset(hyperedge_elements)]
        return False

    def __repr__(self) -> str:
        """Return string representation of hyperedges.

        Returns
        -------
        str
            The __repr__ string representation of the hyperedges.
        """
        return f"ColoredHyperEdgeView({[(tuple(x[0]), x[1]) for x in self]})"

    def __str__(self) -> str:
        """Return string representation of hyperedges.

        Returns
        -------
        str
            The __str__ string representation of the hyperedges.
        """
        return f"ColoredHyperEdgeView({[(tuple(x[0]), x[1]) for x in self]})"

    def skeleton(self, rank: int, store_hyperedge_key: bool = True):
        """Skeleton of the complex.

        Parameters
        ----------
        rank : int
            Rank of the skeleton.
        store_hyperedge_key : bool, default=True
            Whether to return the hyperedge key or not.

        Returns
        -------
        list of frozensets
            The skeleton of rank `rank`.
        """
        if rank not in self.hyperedge_dict:
            return []

        if store_hyperedge_key:
            return sorted(
                [
                    (he, k)
                    for he in self.hyperedge_dict[rank]
                    for k in self.hyperedge_dict[rank][he]
                ]
            )

        return sorted(
            [
                he
                for he in self.hyperedge_dict[rank]
                for k in self.hyperedge_dict[rank][he]
            ]
        )

    def get_rank(self, edge):
        """Get the rank of a given hyperedge.

        Parameters
        ----------
        edge : Iterable, Hashable or ColoredHyperEdge
            The edge for which to get the rank.

        Returns
        -------
        int
            The rank of the given colored hyperedge.
        """
        if isinstance(edge, HyperEdge):
            if len(edge) == 0:
                return 0

            for i in list(self.allranks):
                if frozenset(edge.elements) in self.hyperedge_dict[i]:
                    return i
            raise KeyError(f"hyperedge {edge.elements} is not in the complex")
        if isinstance(edge, str):
            if frozenset({edge}) in self.hyperedge_dict[0]:
                return 0
            raise KeyError(f"hyperedge {frozenset({edge})} is not in the complex")
        if isinstance(edge, Iterable):
            if len(edge) == 0:
                return 0

            for i in list(self.allranks):
                if frozenset(edge) in self.hyperedge_dict[i]:
                    return i
            raise KeyError(f"hyperedge {edge} is not in the complex")
        if isinstance(edge, Hashable) and not isinstance(edge, Iterable):
            if frozenset({edge}) in self.hyperedge_dict[0]:
                return 0
            raise KeyError(f"hyperedge {frozenset({edge})} is not in the complex")
        return None

    @property
    def allranks(self) -> list[int]:
        """All ranks.

        Returns
        -------
        list[int]
            The sorted list of all ranks.
        """
        return sorted(self.hyperedge_dict.keys())


class HyperEdgeView(AtomView):
    """A class for viewing the cells/hyperedges of a combinatorial complex.

    Provides methods for accessing, and retrieving
    information about the cells/hyperedges of a complex.

    Examples
    --------
    >>> hev = tnx.HyperEdgeView()
    """

    def __init__(self) -> None:
        self.hyperedge_dict = {}

    @staticmethod
    def _to_frozen_set(hyperedge):
        """Convert a hyperedge into a frozen set.

        Parameters
        ----------
        hyperedge : HyperEdge | Iterable | Hashable
            The hyperedge that is to be converted to a frozen set.

        Returns
        -------
        frozenset
            Returns a frozenset of the elements contained in the hyperedge.
        """
        if isinstance(hyperedge, HyperEdge):
            hyperedge_ = hyperedge.elements
        elif isinstance(hyperedge, Iterable):
            hyperedge_ = frozenset(hyperedge)
        elif isinstance(hyperedge, Hashable) and not isinstance(hyperedge, Iterable):
            hyperedge_ = frozenset([hyperedge])
        return hyperedge_

    def __getitem__(self, hyperedge: Any) -> dict:
        """Get item.

        Parameters
        ----------
        hyperedge : Hashable or HyperEdge
            DESCRIPTION.

        Returns
        -------
        dict or list or dicts
            Return dict of attributes associated with that hyperedges.
        """
        hyperedge_ = HyperEdgeView._to_frozen_set(hyperedge)
        rank = self.get_rank(hyperedge_)
        return self.hyperedge_dict[rank][hyperedge_]

    @property
    def shape(self) -> tuple[int, ...]:
        """Compute shape.

        Returns
        -------
        tuple[int, ...]
            A tuple representing the shape of the hyperedge.
        """
        return tuple(len(self.hyperedge_dict[i]) for i in self.allranks)

    def __len__(self) -> int:
        """Compute the number of nodes.

        Returns
        -------
        int
            The number of nodes present in the HyperEdgeView.
        """
        return sum(self.shape)

    def __iter__(self) -> Iterator[HyperEdge]:
        """Iterate over the hyperedges.

        Returns
        -------
        Iterator
            Iterator object over the hyperedges.
        """
        return chain.from_iterable(self.hyperedge_dict.values())

    def __contains__(self, atom: Collection) -> bool:
        """Check if e is in the hyperedges.

        Parameters
        ----------
        atom : Collection
            The hyperedge that needs to be checked for containership
            in the HyperEdgeView.

        Returns
        -------
        bool
            Returns `True` if the hyperedge e is contained within the HyperEdgeView,
            else return `False`.
        """
        if len(self.hyperedge_dict) == 0:
            return False
        all_ranks = self.allranks
        if isinstance(atom, HyperEdge):
            if len(atom) == 0:
                return False

            for i in all_ranks:
                if frozenset(atom.elements) in self.hyperedge_dict[i]:
                    return True
            return False
        if isinstance(atom, Iterable):
            if len(atom) == 0:
                return False
            return any(frozenset(atom) in self.hyperedge_dict[i] for i in all_ranks)
        if isinstance(atom, Hashable):
            return frozenset({atom}) in self.hyperedge_dict[0]
        return None

    def __repr__(self) -> str:
        """Return string representation of hyperedges.

        Returns
        -------
        str
            The __repr__ string representation of HyperEdgeView.
        """
        return f"HyperEdgeView({[tuple(x) for x in self]})"

    def __str__(self) -> str:
        """Return string representation of hyperedges.

        Returns
        -------
        str
            The __str__ string representation of HyperEdgeView.
        """
        return f"HyperEdgeView({[tuple(x) for x in self]})"

    def skeleton(
        self,
        rank: int,
        level: Literal[
            "equal",
            "upper",
            "up",
            "lower",
            "down",
            "uppereq",
            "upeq",
            "lowereq",
            "downeq",
        ] = "equal",
    ):
        """Skeleton of the complex.

        Parameters
        ----------
        rank : int
            Rank of the skeleton.
        level : str, default="equal"
            Level of the skeleton.

        Returns
        -------
        list of frozensets
            The skeleton of rank `rank`.
        """
        if level == "equal":
            if rank in self.allranks:
                return sorted(self.hyperedge_dict[rank].keys())
            return []
        if level in {"upper", "up"}:
            elements = []
            for rank_i in self.allranks:
                if rank_i > rank:
                    elements = elements + list(self.hyperedge_dict[rank_i].keys())
            return sorted(elements)
        if level in {"lower", "down"}:
            elements = []
            for rank_i in self.allranks:
                if rank_i < rank:
                    elements = elements + list(self.hyperedge_dict[rank_i].keys())
            return sorted(elements)
        if level in {"uppereq", "upeq"}:
            elements = []
            for rank_i in self.allranks:
                if rank_i >= rank:
                    elements = elements + list(self.hyperedge_dict[rank_i].keys())
            return sorted(elements)
        if level in {"lowereq", "downeq"}:
            elements = []
            for rank_i in self.allranks:
                if rank_i <= rank:
                    elements = elements + list(self.hyperedge_dict[rank_i].keys())
            return sorted(elements)
        raise ValueError(
            "level must be 'equal', 'uppereq', 'lowereq', 'upeq', 'downeq', 'uppereq', 'lower', 'up', or 'down'"
        )

    def get_rank(self, edge):
        """Get the rank of a hyperedge.

        Parameters
        ----------
        edge : Iterable, Hashable or HyperEdge
            The edge for which to get the rank.

        Returns
        -------
        int
            The rank of the given hyperedge.
        """
        if isinstance(edge, HyperEdge):
            if len(edge) == 0:
                return 0

            for i in list(self.allranks):
                if frozenset(edge.elements) in self.hyperedge_dict[i]:
                    return i
            raise KeyError(f"hyperedge {edge.elements} is not in the complex")
        if isinstance(edge, str):
            if frozenset({edge}) in self.hyperedge_dict[0]:
                return 0
            raise KeyError(f"hyperedge {frozenset({edge})} is not in the complex")
        if isinstance(edge, Iterable):
            if len(edge) == 0:
                return 0

            for i in list(self.allranks):
                if frozenset(edge) in self.hyperedge_dict[i]:
                    return i
            raise KeyError(f"hyperedge {edge} is not in the complex")
        if isinstance(edge, Hashable) and not isinstance(edge, Iterable):
            if frozenset({edge}) in self.hyperedge_dict[0]:
                return 0
            raise KeyError(f"hyperedge {frozenset({edge})} is not in the complex")
        return None

    @property
    def allranks(self):
        """All ranks.

        Returns
        -------
        list[hashable]
            The sorted list of all ranks.
        """
        return sorted(self.hyperedge_dict.keys())

    def _get_lower_rank(self, rank):
        """Get a lower rank compared to given rank.

        Parameters
        ----------
        rank : int
            The rank to be used to get a lower rank.

        Returns
        -------
        int
            A rank below the current rank available in the HyperEdgeView.
        """
        if len(self.allranks) == 0:
            return -1

        ranks = sorted(self.allranks)
        if rank <= min(ranks) or rank >= max(ranks):
            return -1
        return ranks[ranks.index(rank) - 1]

    def _get_higher_rank(self, rank):
        """Get a higher rank compared to given rank.

        Parameters
        ----------
        rank : int
            The rank to be used to get a higher rank.

        Returns
        -------
        int
            A rank above the current rank available in the HyperEdgeView.
        """
        if len(self.allranks) == 0:
            return -1
        ranks = sorted(self.allranks)
        if rank <= min(ranks) or rank >= max(ranks):
            return -1
        return ranks[ranks.index(rank) + 1]


class SimplexView(AtomView[Simplex]):
    """Simplex View class.

    The SimplexView class is used to provide a view/read only information
    into a subset of the nodes in a simplex.
    These classes are used in conjunction with the SimplicialComplex class
    for view/read only purposes for simplices in simplicial complexes.

    Parameters
    ----------
    simplex_trie : SimplexTrie
        A SimplexTrie instance containing the simplices in the simplex view.
    """

    def __init__(self, simplex_trie: SimplexTrie) -> None:
        self._simplex_trie = simplex_trie

    def __getitem__(self, simplex: Any) -> dict[Hashable, Any]:
        """Get the dictionary of attributes associated with the given simplex.

        Parameters
        ----------
        simplex : tuple, list or Simplex
            A tuple or list of nodes representing a simplex.

        Returns
        -------
        dict
            A dictionary of attributes associated with the given simplex.

        Raises
        ------
        KeyError
            If the simplex is not in the simplex view.
        """
        if isinstance(simplex, Hashable) and not isinstance(simplex, Iterable):
            simplex = (simplex,)
        else:
            simplex = tuple(sorted(simplex))

        node = self._simplex_trie.find(simplex)
        if node is None:
            raise KeyError(f"Simplex {simplex} is not in the simplex view.")
        return node.attributes

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the number of simplices in each dimension.

        Returns
        -------
        tuple of ints
            A tuple of integers representing the number of simplices in each dimension.
        """
        return tuple(self._simplex_trie.shape)

    def __len__(self) -> int:
        """Return the number of simplices in the SimplexView instance.

        Returns
        -------
        int
            Returns the number of simplices in the SimplexView instance.
        """
        return len(self._simplex_trie)

    def __iter__(self) -> Iterator:
        """Return an iterator over all simplices in the simplex view.

        Returns
        -------
        Iterator
            Returns an iterator over all simplices in the simplex view.
        """
        return iter(node.simplex for node in self._simplex_trie)

    def __contains__(self, atom: Any) -> bool:
        """Check if a simplex is in the simplex view.

        Parameters
        ----------
        atom : Any
            The simplex to be checked for membership in the simplex view.

        Returns
        -------
        bool
            True if the simplex is in the simplex view, False otherwise.

        Examples
        --------
        Check if a node is in the simplex view:

        >>> view = tnx.SimplexView()
        >>> view.faces_dict.append({frozenset({1}): {"weight": 1}})
        >>> view.max_dim = 0
        >>> 1 in view
        True
        >>> 2 in view
        False

        Check if a simplex is in the simplex view:

        >>> view.faces_dict.append({frozenset({1, 2}): {"weight": 1}})
        >>> view.max_dim = 1
        >>> {1, 2} in view
        True
        >>> {1, 3} in view
        False
        >>> {1, 2, 3} in view
        False
        """
        if isinstance(atom, Hashable) and not isinstance(atom, Iterable):
            atom = (atom,)
        else:
            atom = tuple(sorted(atom))
        return atom in self._simplex_trie

    def __repr__(self) -> str:
        """Return string representation that can be used to recreate it.

        Returns
        -------
        str
            Returns the __repr__ representation of the object.
        """
        return f"SimplexView({[tuple(simplex.elements) for simplex in self._simplex_trie]})"

    def __str__(self) -> str:
        """Return detailed string representation of the simplex view.

        Returns
        -------
        str
            Returns the __str__ representation of the object.
        """
        return f"SimplexView({[tuple(simplex.elements) for simplex in self._simplex_trie]})"


class NodeView:
    """Node view class.

    Parameters
    ----------
    nodes : dict[Hashable, Any]
        A dictionary of nodes with their attributes.
    cell_type : type
        The type of the cell.
    colored_nodes : bool, optional
        Whether or not the nodes are colored.
    """

    def __init__(
        self, nodes: dict[Hashable, Any], cell_type, colored_nodes: bool = False
    ) -> None:
        self.nodes = nodes

        if cell_type is None:
            raise ValueError("cell_type cannot be None")

        self.cell_type = cell_type
        self.colored_nodes = colored_nodes

    def __repr__(self) -> str:
        """Return string representation of nodes.

        Returns
        -------
        str
            Returns the __repr__ representation of the object.
        """
        return f"NodeView({list(map(tuple, self.nodes.keys()))})"

    def __iter__(self) -> Iterator:
        """Return an iterator over all nodes in the node view.

        Returns
        -------
        Iterator
            Returns an iterator over all nodes in the node view.
        """
        return iter(self.nodes)

    def __getitem__(self, cell):
        """Get item.

        Parameters
        ----------
        cell : tuple list or AbstractCell or Simplex
            A cell.

        Returns
        -------
        dict or list
            Dict of attributes associated with that cells.
        """
        if isinstance(cell, Iterable):
            cell = frozenset(cell)
            if cell in self.nodes:
                if self.colored_nodes:
                    return self.nodes[cell][0]
                return self.nodes[cell]
        elif isinstance(cell, Hashable):
            if cell in self:
                if self.colored_nodes:
                    return self.nodes[frozenset({cell})][0]
                return self.nodes[frozenset({cell})]

        raise KeyError(f"input {cell} is not in the node set of the complex")

    def __len__(self) -> int:
        """Compute the number of nodes.

        Returns
        -------
        int
            Returns the number of nodes.
        """
        return len(self.nodes)

    def __contains__(self, e) -> bool:
        """Check if e is in the nodes.

        Parameters
        ----------
        e : Hashable | Iterable
            The node to check for.

        Returns
        -------
        bool
            Return `True` if e is contained in NodeView, else Return `False`.
        """
        if isinstance(e, Hashable) and not isinstance(e, self.cell_type):
            return frozenset({e}) in self.nodes
        if isinstance(e, self.cell_type):
            return e.elements in self.nodes
        if isinstance(e, Iterable):
            if len(e) == 1:
                return frozenset(e) in self.nodes
            return None
        return False


class PathView(AtomView[Path]):
    """Path view class."""

    def __init__(self) -> None:
        self.max_dim = -1
        self.faces_dict = []

    def __getitem__(self, path: Any) -> dict:
        """Get the dictionary of attributes associated with the given path.

        Parameters
        ----------
        path : Any
            A tuple or list of nodes representing a path.
            It can also be a Path object.
            It can also be a single node represented by int or str.

        Returns
        -------
        dict or list or dict
            A dictionary of attributes associated with the given path.

        Raises
        ------
        KeyError
            If the path is not in this view.
        """
        if isinstance(path, Path):
            path = path.elements
        if isinstance(path, Hashable) and not isinstance(path, Iterable):
            path = (path,)

        path = tuple(path)
        if path in self.faces_dict[len(path) - 1]:
            return self.faces_dict[len(path) - 1][path]

        raise KeyError(f"input {path} is not in the path dictionary")

    def __iter__(self) -> Iterator:
        """Return an iterator over all paths in the path view.

        Returns
        -------
        Iterator
            Iterator over all paths in the paths view.
        """
        return chain.from_iterable(self.faces_dict)

    def __contains__(self, atom: Any) -> bool:
        """Check if a path is in the path view.

        Parameters
        ----------
        atom : Any
            The path to be checked for membership in the path view.

        Returns
        -------
        bool
            True if the path is in the path view, False otherwise.
        """
        if isinstance(atom, Sequence):
            atom = tuple(atom)
            if not 0 < len(atom) <= self.max_dim + 1:
                return False
            return atom in self.faces_dict[len(atom) - 1]
        if isinstance(atom, Path):
            atom = atom.elements
            if not 0 < len(atom) <= self.max_dim + 1:
                return False
            return atom in self.faces_dict[len(atom) - 1]
        if isinstance(atom, Hashable):
            return (atom,) in self.faces_dict[0]
        return False

    def __len__(self) -> int:
        """Return the number of simplices in the SimplexView instance.

        Returns
        -------
        int
            Returns the number of simplices in the SimplexView instance.
        """
        return sum(self.shape)

    def __repr__(self) -> str:
        """Return string representation that can be used to recreate it.

        Returns
        -------
        str
            Returns the __repr__ representation of the object.
        """
        all_paths: list[tuple[int | str, ...]] = []
        for i in range(len(self.faces_dict)):
            all_paths += [tuple(j) for j in self.faces_dict[i]]

        return f"PathView({all_paths})"

    def __str__(self) -> str:
        """Return detailed string representation of the path view.

        Returns
        -------
        str
            Returns the __str__ representation of the object.
        """
        all_paths: list[tuple[int | str, ...]] = []
        for i in range(len(self.faces_dict)):
            all_paths += [tuple(j) for j in self.faces_dict[i]]

        return f"PathView({all_paths})"

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the number of paths in each dimension.

        Returns
        -------
        tuple of ints
            A tuple of integers representing the number of paths in each dimension.
        """
        return tuple(len(self.faces_dict[i]) for i in range(len(self.faces_dict)))



"""Test Cell class."""

from collections import defaultdict

import numpy as np
import pytest
from scipy.sparse import csr_matrix

from toponetx.classes.cell_complex import CellComplex
from toponetx.utils.structure import (
    compute_set_incidence,
    incidence_to_adjacency,
    neighborhood_list_to_neighborhood_dict,
    sparse_array_to_neighborhood_dict,
    sparse_array_to_neighborhood_list,
)


class TestStructure:
    """Test structure module."""

    def test_sparse_array_to_neighborhood_list(self):
        """Test the sparse_array_to_neighborhood_list function."""
        c = CellComplex()
        c.add_cells_from([[1, 2, 3], [4, 5, 6]], rank=2)
        row, col, B1 = c.incidence_matrix(1, index=True)
        output = sparse_array_to_neighborhood_list(B1)
        expected = [
            (0, 0),
            (1, 0),
            (0, 1),
            (2, 1),
            (1, 2),
            (2, 2),
            (3, 3),
            (4, 3),
            (3, 4),
            (5, 4),
            (4, 5),
            (5, 5),
        ]

        assert list(output) == expected

        output1 = sparse_array_to_neighborhood_list(
            B1, list(col.keys()), list(row.keys())
        )

        expected1 = [
            (1, (1, 2)),
            (2, (1, 2)),
            (1, (1, 3)),
            (3, (1, 3)),
            (2, (2, 3)),
            (3, (2, 3)),
            (4, (4, 5)),
            (5, (4, 5)),
            (4, (4, 6)),
            (6, (4, 6)),
            (5, (5, 6)),
            (6, (5, 6)),
        ]

        assert list(output1) == expected1

        with pytest.raises(ValueError):
            output1 = sparse_array_to_neighborhood_list(B1, list(row.keys()))

    def test_neighborhood_list_to_neighborhood_dict(self):
        """Test the neighborhood_list_to_neighborhood_dict function."""
        neighborhood_list = [
            (0, 0),
            (1, 0),
            (0, 1),
            (2, 1),
            (1, 2),
            (2, 2),
            (3, 3),
            (4, 3),
            (3, 4),
            (5, 4),
            (4, 5),
            (5, 5),
        ]

        out = neighborhood_list_to_neighborhood_dict(neighborhood_list)

        d = defaultdict(
            list, {0: [0, 1], 1: [0, 2], 2: [1, 2], 3: [3, 4], 4: [3, 5], 5: [4, 5]}
        )
        assert out == d

        with pytest.raises(ValueError):
            out = neighborhood_list_to_neighborhood_dict(neighborhood_list, d, None)

    def test_neighborhood_list_to_neighborhood_with_dict(self):
        """Test that neighborhood_list_to_neighborhood_dict works correctly with specified node dictionary."""
        node_dict = {1: "a", 2: "b", 3: "c"}
        neigh_list = [(1, 2), (2, 1), (2, 3), (3, 2)]
        neigh_dict = neighborhood_list_to_neighborhood_dict(
            neigh_list, node_dict, node_dict
        )
        assert len(neigh_dict) == 3
        assert neigh_dict["a"] == ["b"]
        assert set(neigh_dict["b"]) == {"a", "c"}  # order irrelevant
        assert neigh_dict["c"] == ["b"]

    def test_sparse_array_to_neighborhood_dict(self):
        """Test the sparse_array_to_neighborhood_dict function."""
        c = CellComplex()
        c.add_cells_from([[1, 2, 3], [4, 5, 6]], rank=2)
        _, _, B1 = c.incidence_matrix(1, index=True)
        output = sparse_array_to_neighborhood_dict(B1)

        d = defaultdict(
            list, {0: [0, 1], 1: [0, 2], 2: [1, 2], 3: [3, 4], 4: [3, 5], 5: [4, 5]}
        )
        assert output == d

    def test_incidence_to_adjacency(self):
        """Test incidence to adjacency.

        uses transposed of cell-edge incidence for cell complex [(1,2,3,4), (1,2,4,3)]
        """
        incidence = csr_matrix(
            [[1, 0, -1, 1, 0, 1], [1, -1, 0, 0, 1, -1]]
        )  # already transposed (check for upper adj.)
        adj = incidence_to_adjacency(incidence)
        expected_adj = csr_matrix(
            [
                [
                    0,
                    1,
                    1,
                    1,
                    1,
                    2,
                ],  # (1,2) and (3,4) are upper-adjacent in both 2-cells
                [1, 0, 0, 0, 1, 1],
                [1, 0, 0, 1, 0, 1],
                [1, 0, 1, 0, 0, 1],
                [1, 1, 0, 0, 0, 1],
                [2, 1, 1, 1, 1, 0],
            ]
        )
        assert (
            (adj != expected_adj).nnz == 0
        )  # tests sparsity of difference -> if the difference has no non-zero entries, it is the same

    def test_compute_set_incidence(self):
        """Test compute_set_incidence."""
        # Basic functionality
        children = [1, 2, 3]
        uidset = [10, 11, 12]
        result = compute_set_incidence(children, uidset)
        assert np.array_equal(
            result.toarray(), np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
        )

        # Sparse matrix and index
        children = [1, 2]
        uidset = [10, 12]
        result_ndict, result_edict, result_mp = compute_set_incidence(
            children, uidset, sparse=True, index=True
        )
        assert isinstance(result_mp, csr_matrix)
        assert result_mp.shape == (len(children), len(uidset))
        assert len(result_ndict) == len(children)
        assert len(result_edict) == len(uidset)

        # Non-sparse matrix
        children = [1, 2, 3]
        uidset = [10, 11, 12]
        result_mp = compute_set_incidence(children, uidset, sparse=False)
        assert isinstance(result_mp, np.ndarray)
        assert result_mp.shape == (len(children), len(uidset))

        # Empty input without index
        children = []
        uidset = []
        result = compute_set_incidence(children, uidset)
        assert result.size == 1

        # Empty input with index
        children = []
        uidset = []
        result_ndict, result_edict, result_mp = compute_set_incidence(
            children, uidset, index=True
        )
        assert result.size == 1



"""Test components module."""

import pytest

from toponetx.algorithms.components import (
    connected_component_subcomplexes,
    connected_components,
    s_component_subcomplexes,
    s_connected_components,
)
from toponetx.classes.cell_complex import CellComplex
from toponetx.classes.simplicial_complex import SimplicialComplex


class TestComponents:
    """Test components module."""

    def test_s_connected_components(self):
        """Test_s_connected_components."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([5, 6, 7], rank=2)

        # Test the function with cells=True
        result = list(s_connected_components(CC, s=1, cells=True))
        expected_result = [
            {(2, 3), (2, 3, 4), (2, 4), (3, 4)},
            {(5, 6), (5, 6, 7), (5, 7), (6, 7)},
        ]
        assert result == expected_result

        # Test the function with cells=False
        result = list(s_connected_components(CC, s=1, cells=False))
        expected_result = [{2, 3, 4}, {5, 6, 7}]
        assert result == expected_result

        # test with ColoredHyperGraph
        CHG = CC.to_colored_hypergraph()
        result = list(s_connected_components(CHG, s=1, cells=True))
        expected_result = [
            {
                (frozenset({3, 4}), 0),
                (frozenset({2, 4}), 0),
                (frozenset({2, 3}), 0),
                (frozenset({2, 3, 4}), 0),
            },
            {
                (frozenset({5, 7}), 0),
                (frozenset({5, 6, 7}), 0),
                (frozenset({5, 6}), 0),
                (frozenset({6, 7}), 0),
            },
        ]
        assert result == expected_result

        result = list(s_connected_components(CHG, s=1, cells=False))
        expected_result = [{2, 3, 4}, {5, 6, 7}]
        assert result == expected_result

        # test cells=True, return_singletons=False
        CC.add_node(10)
        result = list(s_connected_components(CC, s=1, cells=True))
        expected_result = [
            {(2, 3), (2, 3, 4), (2, 4), (3, 4)},
            {(5, 6), (5, 6, 7), (5, 7), (6, 7)},
        ]
        assert result == expected_result

        # Test the function with return_singletons=True
        CC2 = CellComplex()
        CC2.add_cell([1, 2, 3, 4], rank=2)
        CC2.add_cell([1, 2, 3, 4], rank=2)
        CC2.add_cell([1, 2, 3, 4], rank=2)
        CC2.add_node(0)
        CC2.add_node(10)

        result = list(
            s_connected_components(CC2, s=1, cells=False, return_singletons=True)
        )
        expected_result = [{0}, {1, 2, 3, 4}, {10}]

        # Test the function with return_singletons=False
        result = list(
            s_connected_components(CC2, s=2, cells=False, return_singletons=False)
        )
        expected_result = [{1, 2, 3, 4}]
        assert result == expected_result

        # invalid input
        with pytest.raises(TypeError):
            _ = list(s_connected_components(SimplicialComplex()))

    def test_s_component_subcomplexes(self):
        """Test_s_component_subcomplexes."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([5, 6, 7], rank=2)

        # Test the function with cells=True
        result = list(s_component_subcomplexes(CC, s=1, cells=True))
        expected_result = [CellComplex(), CellComplex()]
        expected_result[0].add_cell([2, 3, 4], rank=2)
        expected_result[1].add_cell([5, 6, 7], rank=2)
        assert len(result[0].cells) == len(expected_result[0].cells)
        assert len(result[1].cells) == len(expected_result[1].cells)

        CC = CellComplex()
        # Test the function with return_singletons=False
        result = list(
            s_component_subcomplexes(CC, s=1, cells=False, return_singletons=False)
        )
        expected_result = []
        assert result == expected_result

    def test_connected_components(self):
        """Test_connected_components."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([5, 6, 7], rank=2)

        # Test the function with cells=False and return_singletons=True
        result = list(connected_components(CC, cells=True, return_singletons=True))
        expected_result = [
            {(2, 3), (2, 3, 4), (2, 4), (3, 4)},
            {(5, 6), (5, 6, 7), (5, 7), (6, 7)},
        ]
        assert result == expected_result

        CC = CellComplex()
        # Test the function with cells=False and return_singletons=False
        result = list(connected_components(CC, cells=False, return_singletons=False))
        expected_result = []
        assert result == expected_result

    def test_connected_component_subcomplexes(self):
        """Test_connected_component_subcomplexes."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([5, 6, 7], rank=2)

        # Test the function with return_singletons=True
        result = list(connected_component_subcomplexes(CC, return_singletons=True))
        expected_result = [CellComplex(), CellComplex()]
        expected_result[0].add_cell([2, 3, 4], rank=2)
        expected_result[1].add_cell([5, 6, 7], rank=2)
        assert len(result[0].cells) == len(expected_result[0].cells)
        assert len(result[1].cells) == len(expected_result[1].cells)

        # Test the function with return_singletons=False
        CC = CellComplex()  # Initialize your class object
        result = list(connected_component_subcomplexes(CC, return_singletons=False))
        expected_result = []
        assert result == expected_result



"""Generators for random simplicial complexes."""

import random
from collections.abc import Sequence
from itertools import combinations

import networkx as nx

from toponetx.classes import SimplicialComplex
from toponetx.transform import graph_to_clique_complex

__all__ = [
    "linial_meshulam_complex",
    "multiparameter_linial_meshulam_complex",
    "random_clique_complex",
]


def linial_meshulam_complex(n: int, p: float, seed=None) -> SimplicialComplex:
    """Generate a random Linial-Meshulam simplicial complex [1]_.

    The random simplicial complex is generated by starting with a complete graph (i.e.,
    a complete 1-skeleton) and then adding 2-simplices with probability `p`.

    Parameters
    ----------
    n : int
        The number of nodes.
    p : float
        The probability of a 2-simplex being added.
    seed : int or random_state (optional)
        Indicator of random number generation state.

    Returns
    -------
    SimplicialComplex
        A random Linial-Meshulam simplicial complex.

    Notes
    -----
    This is equivalent to `multiparameter_linial_meshulam_complex` with edge
    probability one and 2-simplex probability `p`.

    References
    ----------
    .. [1] Nathan Linial and Roy Meshulam, "Homological Connectivity Of Random
           2-Complexes," Combinatorica 26, no. 4 (August 2006): 475-87,
           https://doi.org/10.1007/s00493-006-0027-9.
    """
    return multiparameter_linial_meshulam_complex(n, [1.0, p], seed)


def random_clique_complex(n: int, p: float, seed=None) -> SimplicialComplex:
    """Generate a random clique complex [1]_.

    In random clique complexes, a sample from an Erdős-Rényi graph is drawn and all
    cliques are identified with a face of a simplicial complex.

    Parameters
    ----------
    n : int
        The number of nodes.
    p : float
        The probability of an edge between two nodes.
    seed : int or random_state (optional)
        Indicator of random number generation state.

    Returns
    -------
    SimplicialComplex
        A random clique complex.

    References
    ----------
    .. [1] Matthew Kahle, "Topology of Random Clique Complexes,"
           Discrete Mathematics 309, no. 6 (April 2009): 1658-71,
           https://doi.org/10.1016/j.disc.2008.02.037.
    """
    graph = nx.gnp_random_graph(n, p, seed)
    return graph_to_clique_complex(graph)


def multiparameter_linial_meshulam_complex(
    n: int, ps: Sequence[float], seed: random.Random | None = None
) -> SimplicialComplex:
    """Generate a random multiparameter Linial-Meshulam complex.

    The random simplicial complex is generated inductively, starting with a set of `n`
    nodes. The probability for two nodes to be connected is given by `ps[0]`. Then, the
    probability for each `k`-clique to be identified by a `k-1`-simplex is given by
    `ps[k-2]`.

    Parameters
    ----------
    n : int
        The number of nodes.
    ps : sequence of floats
        The probabilities for each simplex rank. The first entry is the probability for
        two nodes to be connected. The second entry dictates the probability for a
        2-simplex to be added to the previous 1-skeleton etc.
    seed : int or random_state (optional)
        Indicator of random number generation state.

    Returns
    -------
    SimplicialComplex
        A random multiparameter Linial-Meshulam complex.

    Notes
    -----
    For `p[0] = 1.0` and `p[1] = p`, this recovers the standard random Linial-Meshulam
    complex provided by `random_linial_meshulam_complex`.
    """
    rng = nx.utils.create_py_random_state(seed)

    graph = nx.gnp_random_graph(n, ps[0], rng)
    SC = SimplicialComplex()

    for clique in nx.enumerate_all_cliques(graph):
        if len(clique) <= 2:
            SC.add_simplex(clique)
        elif len(clique) <= len(ps) + 1:
            # By random choice, we may have not taken all (k-1)-simplices of the
            # current k-clique. If that is the case, ignore this clique.
            simplices = SC.simplices
            if any(s not in simplices for s in combinations(clique, len(clique) - 1)):
                continue
            if rng.random() < ps[len(clique) - 2]:
                SC.add_simplex(clique)
        else:
            # `nx.enumerate_all_cliques` returns cliques in ascending order of size.
            # Abort calling the generator once we reach cliques for which we do not
            # have a probability.
            break

    return SC



"""Read and write complexes as a list of their atoms."""

from ast import literal_eval
from collections.abc import Generator, Hashable, Iterable
from itertools import combinations
from typing import Literal, overload

import networkx as nx

from toponetx.classes import CellComplex, SimplicialComplex

__all__ = [
    "generate_atomlist",
    "load_from_atomlist",
    "parse_atomlist",
    "write_atomlist",
]


def _atomlist_line(atom: Iterable[Hashable] | Hashable, attributes: dict) -> str:
    """Construct a single line of an atom list.

    Parameters
    ----------
    atom : iterable of hashable or hashable
        The atom to write.
    attributes : dict
        Attributes associated with the atom.

    Returns
    -------
    str
        The line of the atom list that represents the given atom.
    """
    line = " ".join(map(str, atom)) if isinstance(atom, Iterable) else str(atom)

    if len(attributes) > 0:
        line += " " + str(attributes)

    return line


def _generate_atomlist_simplicial(
    domain: SimplicialComplex,
) -> Generator[str, None, None]:
    """Generate an atom list from a simplicial complex.

    The list of atoms is truncated to only contain maximal simplices and simplices with user-defined attributes. All
    other simplices are implicitly contained by the simplex property already.

    Parameters
    ----------
    domain : SimplicialComplex
        The simplicial complex to be converted to an atom list.

    Yields
    ------
    str
        One line of the atom list, which corresponds to one atom of the complex together with its attributes.
    """
    for atom in domain.simplices:
        data = domain[atom].copy()
        data.pop("is_maximal", None)
        data.pop("membership", None)

        if len(data) == 0 and not domain.is_maximal(atom):
            continue

        yield _atomlist_line(atom, data)


def _generate_atomlist_cell(domain: CellComplex) -> Generator[str, None, None]:
    """Generate an atom list from a cell complex.

    The list of atoms is truncated to only contain maximal cells and cells with user-defined attributes. All
    other cells are implicitly contained already.
    We add a special `rank` attribute to cells of cardinality 2 that have rank 2 to differentiate them from edges.

    Parameters
    ----------
    domain : CellComplex
        The cell complex to be converted to an atom list.

    Yields
    ------
    str
        One line of the atom list, which corresponds to one atom of the complex together with its attributes.
    """
    for atom in domain.nodes:
        if len(domain.neighbors(atom)) == 0 or len(domain._G.nodes[atom]) > 0:
            yield _atomlist_line(atom, domain._G.nodes[atom])

    covered_edges = set()
    for cell in domain.cells:
        for edge in combinations(cell, 2):
            covered_edges.add(tuple(sorted(edge)))
    for atom in domain.edges:
        if len(domain._G.edges[atom]) > 0 or tuple(sorted(atom)) not in covered_edges:
            yield _atomlist_line(atom, domain._G.edges[atom])

    for atom in domain.cells:
        attributes = atom._attributes.copy()
        if len(atom) == 2:
            attributes["rank"] = 2
        yield _atomlist_line(atom, attributes)


def generate_atomlist(
    domain: CellComplex | SimplicialComplex,
) -> Generator[str, None, None]:
    """Generate an atom list from a complex.

    The list of atoms is truncated to only contain maximal atoms and atoms with user-defined attributes. All
    other atoms are implicitly contained already.
    For cell complexes, e add a special `rank` attribute to cells of cardinality 2 that have rank 2 to differentiate
    them from edges.

    Parameters
    ----------
    domain : CellComplex or SimplicialComplex
        The complex to be converted to an atom list.

    Yields
    ------
    str
        One line of the atom list, which corresponds to one atom of the complex together with its attributes.

    Examples
    --------
    Generate a list of atoms from a simplicial complex:

    >>> SC = tnx.SimplicialComplex()
    >>> SC.add_simplex((1,), weight=1.0)
    >>> SC.add_simplex((1, 2, 3), weight=4.0)
    >>> list(tnx.generate_atomlist(SC))
    ["1 {'weight': 1.0}", "1 2 3 {'weight': 4.0}"]

    Generate a list of atoms from a cell complex:

    >>> CC = tnx.CellComplex()
    >>> CC.add_cell((1, 2, 3), rank=2, weight=4.0)
    >>> list(tnx.generate_atomlist(CC))
    ["1 2 3 {'weight': 4.0}"]
    """
    if isinstance(domain, SimplicialComplex):
        yield from _generate_atomlist_simplicial(domain)
    elif isinstance(domain, CellComplex):
        yield from _generate_atomlist_cell(domain)
    else:
        raise TypeError(f"Expected a cell or simplicial complex, got {type(domain)}.")


@nx.utils.open_file(1, "wb")
def write_atomlist(
    domain: CellComplex | SimplicialComplex, path, encoding="utf-8"
) -> None:
    """Write an atom list to a file.

    Parameters
    ----------
    domain : CellComplex or SimplicialComplex
        The complex to be converted to an atom list.
    path : file or str
        File or filename to write. If a file is provided, it must be opened in `wb`
        mode. Filenames ending in .gz or .bz2 will be compressed.
    encoding : str, default="utf-8"
       Specify which encoding to use when writing file.

    Raises
    ------
    TypeError
        If the domain is not a cell or simplicial complex.
    """
    if not isinstance(domain, CellComplex | SimplicialComplex):
        raise TypeError(f"Expected a cell or simplicial complex, got {type(domain)}.")

    for line in generate_atomlist(domain):
        line += "\n"
        path.write(line.encode(encoding))


@overload
def load_from_atomlist(
    filepath: str, complex_type: Literal["cell"], nodetype=None, encoding="utf-8"
) -> CellComplex:  # numpydoc ignore=GL08
    pass


@overload
def load_from_atomlist(
    filepath: str, complex_type: Literal["simplicial"], nodetype=None, encoding="utf-8"
) -> SimplicialComplex:  # numpydoc ignore=GL08
    pass


@nx.utils.open_file(0, "rb")
def load_from_atomlist(
    path, complex_type: Literal["cell", "simplicial"], nodetype=None, encoding="utf-8"
) -> CellComplex | SimplicialComplex:
    """Load a complex from an atom list.

    Parameters
    ----------
    path : file or str
        File or filename to read. If a file is provided, it must be opened in `rb`
        mode. Filenames ending in .gz or .bz2 will be uncompressed.
    complex_type : {"cell", "simplicial"}
        The type of complex that should be constructed based on the atom list.
    nodetype : callable, optional
        Convert node data from strings to the specified type.
    encoding : str, default="utf-8"
         Specify which encoding to use when reading file.

    Returns
    -------
    CellComplex or SimplicialComplex
        The complex that was loaded from the atom list.

    Raises
    ------
    ValueError
        If the complex type is unknown.
    """
    return parse_atomlist(
        (line.decode(encoding) for line in path), complex_type, nodetype
    )


@overload
def parse_atomlist(
    lines: Iterable[str], complex_type: Literal["cell"], nodetype=None
) -> CellComplex:  # numpydoc ignore=GL08
    pass


@overload
def parse_atomlist(
    lines: Iterable[str], complex_type: Literal["simplicial"], nodetype=None
) -> SimplicialComplex:  # numpydoc ignore=GL08
    pass


def parse_atomlist(
    lines: Iterable[str], complex_type: Literal["cell", "simplicial"], nodetype=None
) -> CellComplex | SimplicialComplex:
    """Parse an atom list.

    Parameters
    ----------
    lines : iterable of str
        List of lines.
    complex_type : {"cell", "simplicial"}
        Complex type.
    nodetype : callable, optional
        Node type.

    Returns
    -------
    CellComplex or SimplicialComplex
        The complex that was parsed from the atom list.

    Raises
    ------
    ValueError
        If the complex type is unknown.
    """
    domain: CellComplex | SimplicialComplex
    if complex_type == "cell":
        domain = CellComplex()
    elif complex_type == "simplicial":
        domain = SimplicialComplex()
    else:
        raise ValueError(f"Unknown complex type {complex_type}.")

    for line in lines:
        attributes_pos = line.find("{")
        if attributes_pos == -1:
            elements_str = line
            attributes = {}
        else:
            elements_str = line[:attributes_pos].strip()
            attributes = literal_eval(line[attributes_pos:])

        elements = elements_str.split(" ")
        elements = [e.strip() for e in elements]
        if nodetype is not None:
            elements = [nodetype(e) for e in elements]

        if isinstance(domain, CellComplex):
            if "rank" in attributes:
                rank = attributes.pop("rank")
            else:
                rank = min(len(elements) - 1, 2)

            if rank == 0:
                domain.add_node(elements[0], **attributes)
            else:
                domain.add_cell(elements, rank=rank, **attributes)
        elif isinstance(domain, SimplicialComplex):
            domain.add_simplex(elements, **attributes)

    return domain



"""Simplex Class."""

from collections.abc import Collection, Hashable, Iterable
from functools import total_ordering
from typing import Self

from toponetx.classes.complex import Atom
from toponetx.utils.iterable import is_ordered_subset

__all__ = ["Simplex"]


@total_ordering
class Simplex[ElementType: Hashable](Atom):
    """A class representing a simplex in a simplicial complex.

    This class represents a simplex in a simplicial complex, which is a set of nodes with a specific dimension. The
    simplex is immutable, and the nodes in the simplex must be hashable and unique.

    Parameters
    ----------
    elements : Collection[Hashable]
        The nodes in the simplex.
    **kwargs : keyword arguments, optional
        Additional attributes to be associated with the simplex.

    Examples
    --------
    >>> # Create a 0-dimensional simplex (point)
    >>> s = tnx.Simplex((1,))
    >>> # Create a 1-dimensional simplex (line segment)
    >>> s = tnx.Simplex((1, 2))
    >>> # Create a 2-dimensional simplex (triangle)
    >>> simplex1 = tnx.Simplex((1, 2, 3))
    >>> simplex2 = tnx.Simplex(("a", "b", "c"))
    >>> # Create a 3-dimensional simplex (tetrahedron)
    >>> simplex3 = tnx.Simplex((1, 2, 4, 5), weight=1)
    """

    elements: frozenset[Hashable]

    def __init__(
        self,
        elements: Collection[ElementType],
        **kwargs,
    ) -> None:
        self.validate_attributes(kwargs)

        if len(elements) != len(set(elements)):
            raise ValueError("A simplex cannot contain duplicate nodes.")

        super().__init__(tuple(sorted(elements)), **kwargs)

    def __contains__(self, item: ElementType | Iterable[ElementType]) -> bool:
        """Return True if the given element is a subset of the nodes.

        Parameters
        ----------
        item : Any
            The element to be checked.

        Returns
        -------
        bool
            True if the given element is a subset of the nodes.

        Examples
        --------
        >>> s = tnx.Simplex((1, 2, 3))
        >>> 1 in s
        True
        >>> 4 in s
        False
        >>> (1, 2) in s
        True
        >>> (1, 4) in s
        False
        """
        if isinstance(item, Iterable):
            item = tuple(sorted(item))
            return is_ordered_subset(item, self.elements)
        return super().__contains__(item)

    def __le__(self, other) -> bool:
        """Return True if this simplex comes before the other simplex in the lexicographic order.

        Parameters
        ----------
        other : Any
            The other simplex to compare with.

        Returns
        -------
        bool
            True if this simplex comes before the other simplex in the lexicographic order.
        """
        if not isinstance(other, Simplex):
            return NotImplemented
        return tuple(self.elements) <= tuple(other.elements)

    @staticmethod
    def validate_attributes(attributes: dict) -> None:
        """Validate the attributes of the simplex.

        Parameters
        ----------
        attributes : dict
            The attributes to be validated.

        Raises
        ------
        ValueError
            If the attributes contain the reserved keys `is_maximal` or `membership`.
        """
        if "is_maximal" in attributes or "membership" in attributes:
            raise ValueError(
                "Special attributes `is_maximal` and `membership` are reserved."
            )

    def sign(self, face: "Simplex[ElementType]") -> int:
        """Calculate the sign of the simplex with respect to a given face.

        Parameters
        ----------
        face : Simplex
            A face of the simplex.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        """Return string representation of the simplex.

        Returns
        -------
        str
            A string representation of the simplex.
        """
        return f"Simplex({tuple(self.elements)})"

    def __str__(self) -> str:
        """Return string representation of the simplex.

        Returns
        -------
        str
            A string representation of the simplex.
        """
        return f"Nodes set: {tuple(self.elements)}, attrs: {self._attributes}"

    def clone(self) -> Self:
        """Return a copy of the simplex.

        The clone method by default returns an independent shallow copy of the simplex
        and attributes. That is, if an attribute is a container, that container is
        shared by the original and the copy. Use Python's `copy.deepcopy` for new
        containers.

        Returns
        -------
        Simplex
            A copy of this simplex.
        """
        return self.__class__(self.elements, **self._attributes)



"""Utils loading datasets."""

import networkx as nx
import numpy as np
import pandas as pd
import requests


def load_ppi():
    """Load the protein-protein-interaction graph.

    In the graph, high interaction score is represented as low weight.

    Returns
    -------
    networkx.Graph
        Graph with nodes that are proteins and edges corresponding to their
        chemical interactions.

    References
    ----------
    https://towardsdatascience.com/visualizing-protein-networks-in-python-58a9b51be9d5
    """
    protein_list = [
        "TPH1",
        "COMT",
        "SLC18A2",
        "HTR1B",
        "HTR2C",
        "HTR2A",
        "MAOA",
        "TPH2",
        "HTR1A",
        "HTR7",
        "SLC6A4",
        "GABBR2",
        "POMC",
        "GNAI3",
        "NPY",
        "ADCY1",
        "PDYN",
        "GRM2",
        "GRM3",
        "GABBR1",
    ]
    proteins = "%0d".join(protein_list)
    url = f"https://string-db.org/api/tsv/network?identifiers={proteins}&species=9606"
    r = requests.get(url, timeout=10)

    lines = r.text.split("\n")
    data = [line.split("\t") for line in lines]
    protein_df = pd.DataFrame(data[1:-1], columns=data[0])

    interactions = protein_df[["preferredName_A", "preferredName_B", "score"]]

    G = nx.Graph(name="Protein Interaction Graph")
    interactions = np.array(interactions)
    for interaction in interactions:
        protein_a = interaction[0]
        protein_b = interaction[1]
        interaction_weight = float(interaction[2])
        G.add_weighted_edges_from([(protein_a, protein_b, interaction_weight)])
    return G



"""Test ReportViews module."""

import pytest

from toponetx.classes import (
    Cell,
    CellComplex,
    ColoredHyperGraph,
    CombinatorialComplex,
    HyperEdge,
    Path,
    PathComplex,
    Simplex,
    SimplicialComplex,
)
from toponetx.classes.reportviews import (
    HyperEdgeView,
    NodeView,
)


class TestReportViews_CellView:
    """Test Cell View Class of the ReportViews module."""

    CX = CellComplex()
    CX._insert_cell((1, 2, 3, 4), color="red")
    CX._insert_cell((1, 2, 3, 4), shape="square")
    CX._insert_cell((2, 3, 4, 5))

    CV = CX._cells
    cell_1 = Cell((1, 2, 3, 4))
    cell_2 = Cell((1, 2, 3, 8))

    def test_cell_view_report_getitem_method(self):
        """Test the __getitem__ method of CellView class."""
        with pytest.raises(KeyError, match="not in the cell dictionary"):
            self.CV.__getitem__(self.cell_2)

        assert self.CV.__getitem__(self.cell_1) == {"shape": "square"}

        with pytest.raises(KeyError, match="not in the cell dictionary"):
            self.CV.__getitem__((1, 2, 3, 8))

        with pytest.raises(TypeError):
            self.CV.__getitem__(1)

    def test_cell_view_raw_type_error(self):
        """Test the raw method for the CellView class."""
        with pytest.raises(TypeError):
            _ = self.CV.raw(1)

        with pytest.raises(TypeError):
            self.CV.raw("testing string")

        with pytest.raises(TypeError):
            self.CV.raw({1})

        with pytest.raises(TypeError):
            self.CV.raw({"1": "some_string"})

    def test_cell_view_raw_key_error(self):
        """Test the raw method for the CellView class."""
        # Raw Method should raise a KeyError exception when cell is not present
        with pytest.raises(KeyError, match="not in the cell dictionary"):
            self.CV.raw(self.cell_2)

    def test_cell_view_raw_method(self):
        """Test the raw method for the CellView class."""
        assert self.CV.raw(self.cell_1) == Cell((1, 2, 3, 4))

    def test_cell_view_raw_method_8(self):
        """Test the raw method for the CellView class."""
        with pytest.raises(KeyError) as exp_exception:
            self.CV.raw((1, 2, 3, 8))

        assert (
            str(exp_exception.value)
            == "'cell (1, 2, 3, 8) is not in the cell dictionary'"
        )

    def test_cell_view_contains_method(self):
        """Test the contains method for the CellView class."""
        CX = CellComplex()
        # Inserting some cells in CX
        CX._insert_cell((1, 2, 3, 4), color="red")
        CX._insert_cell((1, 2, 3, 4), color="green")
        CX._insert_cell((7, 8, 9, 1), shape="square")
        CX._insert_cell((2, 3, 4, 5))

        # Creating Cell View from Cell Complex
        CV = CX._cells

        # To test we are using three cells that are present in the cell complex
        # these are the three cases of multiple homotopic cells present (with properties)
        # and single cell (with and without properties attribute)

        cell_1 = Cell((1, 2, 3, 4))
        cell_2 = Cell((2, 3, 4, 1))
        cell_3 = Cell((1, 3, 2, 4))
        cell_4 = Cell((7, 8, 9, 1))
        cell_5 = Cell((1, 7, 8, 9))
        cell_6 = Cell((7, 1, 8, 9))
        cell_7 = Cell((2, 3, 4, 5))
        cell_8 = Cell((3, 4, 5, 2))
        cell_9 = Cell((3, 2, 4, 5))
        cell_10 = Cell((1, 2, 3, 8))

        assert CV.__contains__(cell_1) is True
        assert CV.__contains__(cell_2) is True
        assert CV.__contains__(cell_3) is False

        assert CV.__contains__(cell_4) is True
        assert CV.__contains__(cell_5) is True
        assert CV.__contains__(cell_6) is False

        assert CV.__contains__(cell_7) is True
        assert CV.__contains__(cell_8) is True
        assert CV.__contains__(cell_9) is False

        assert CV.__contains__(cell_10) is False

        # Now testing the same with list and tuples as inputs

        cell_1_tp = (1, 2, 3, 4)
        cell_2_tp = (2, 3, 4, 1)
        cell_3_tp = (1, 3, 2, 4)
        cell_4_tp = (7, 8, 9, 1)
        cell_5_tp = (1, 7, 8, 9)
        cell_6_tp = (7, 1, 8, 9)
        cell_7_tp = (2, 3, 4, 5)
        cell_8_tp = (3, 4, 5, 2)
        cell_9_tp = (3, 2, 4, 5)
        cell_10_tp = (1, 2, 3, 8)

        assert CV.__contains__(cell_1_tp) is True
        assert CV.__contains__(cell_2_tp) is True
        assert CV.__contains__(cell_3_tp) is False

        assert CV.__contains__(cell_4_tp) is True
        assert CV.__contains__(cell_5_tp) is True
        assert CV.__contains__(cell_6_tp) is False

        assert CV.__contains__(cell_7_tp) is True
        assert CV.__contains__(cell_8_tp) is True
        assert CV.__contains__(cell_9_tp) is False

        assert CV.__contains__(cell_10_tp) is False

        assert CV.__contains__(list(cell_1_tp)) is True
        assert CV.__contains__(list(cell_2_tp)) is True
        assert CV.__contains__(list(cell_3_tp)) is False

        assert CV.__contains__(list(cell_4_tp)) is True
        assert CV.__contains__(list(cell_5_tp)) is True
        assert CV.__contains__(list(cell_6_tp)) is False

        assert CV.__contains__(list(cell_7_tp)) is True
        assert CV.__contains__(list(cell_8_tp)) is True
        assert CV.__contains__(list(cell_9_tp)) is False

        assert CV.__contains__(list(cell_10_tp)) is False

        assert CV.__contains__({1, 2, 3, 4}) is False
        assert CV.__contains__(1) is False
        assert CV.__contains__("1234") is False
        assert CV.__contains__({"1": True, "2": True, "3": True, "4": True}) is False

    def test_cell_view_repr_method(self):
        """Test the __repr__ method for the CellView class."""
        assert (
            self.CV.__repr__() == "CellView([Cell((1, 2, 3, 4)), Cell((2, 3, 4, 5))])"
        )

    def test_cell_view_str_method(self):
        """Test the __str__ method for the CellView class."""
        assert self.CV.__str__() == "CellView([Cell((1, 2, 3, 4)), Cell((2, 3, 4, 5))])"


class TestReportViews_HyperEdgeView:
    """Test the HyperEdgeView class of the ReportView class."""

    def test_hyper_edge_view_to_frozenset(self):
        """Testing the _to_frozen_set static method of the Hyperedge View Class."""
        he_1 = HyperEdge((1, 2, 3, 4), rank=2)
        he_2 = HyperEdge({1, 2, 3, 4}, rank=2)
        he_3 = [1, 2, 3, 4]
        he_4 = (1, 2, 3, 4)
        he_5 = 1

        assert HyperEdgeView._to_frozen_set(he_1) == frozenset({1, 2, 3, 4})
        assert HyperEdgeView._to_frozen_set(he_2) == frozenset({1, 2, 3, 4})
        assert HyperEdgeView._to_frozen_set(he_3) == frozenset({1, 2, 3, 4})
        assert HyperEdgeView._to_frozen_set(he_4) == frozenset({1, 2, 3, 4})
        assert HyperEdgeView._to_frozen_set(he_5) == frozenset({1})

    def test_hyper_edge_view_contains(self):
        """Test the __contains__ method for the Hyperedge View Class."""
        hev = HyperEdgeView()

        he_1 = HyperEdge((1, 2, 3, 4), rank=2)
        he_2 = HyperEdge({1, 2, 3, 5}, rank=2)
        he_3 = [1, 2, 3, 4]
        he_4 = (1, 2, 3, 5)
        he_5 = []
        he_6 = HyperEdge({}, rank=1)

        CC = CombinatorialComplex([he_1], name="trial_CC")
        hev_2 = CC._complex_set

        assert hev.__contains__(he_1) is False
        assert hev.__contains__(he_5) is False
        assert hev.__contains__(he_6) is False

        assert hev_2.__contains__(he_1) is True
        assert hev_2.__contains__(he_2) is False
        assert hev_2.__contains__(he_3) is True
        assert hev_2.__contains__(he_4) is False
        assert hev_2.__contains__(he_5) is False
        assert hev_2.__contains__(he_6) is False

    def test_hyper_edge_view_repr(self):
        """Test the __repr__ method for the Hyperedge View Class."""
        he_1 = HyperEdge((1, 2, 3, 4), rank=2)

        CC = CombinatorialComplex([he_1], name="trial_CC")

        assert (
            CC._complex_set.__repr__()
            == "HyperEdgeView([(1, 2, 3, 4), (1,), (2,), (3,), (4,)])"
        )

    def test_hyper_edge_view_str(self):
        """Test the __str__ method for the Hyperedge View Class."""
        he_1 = HyperEdge((1, 2, 3, 4), rank=2)

        CC = CombinatorialComplex([he_1], name="trial_CC")

        assert (
            CC._complex_set.__str__()
            == "HyperEdgeView([(1, 2, 3, 4), (1,), (2,), (3,), (4,)])"
        )

    def test_hyper_edge_view_skeleton(self):
        """Test the skeleton method for the Hyperedge View Class."""
        he_1 = HyperEdge((1, 2, 3, 4), rank=2)

        CC = CombinatorialComplex([he_1], name="trial_CC")
        hev = CC._complex_set

        assert hev.skeleton(rank=0) == [
            frozenset({1}),
            frozenset({2}),
            frozenset({3}),
            frozenset({4}),
        ]

        assert hev.skeleton(rank=2) == [frozenset({1, 2, 3, 4})]

        assert hev.skeleton(rank=3) == []

        assert hev.skeleton(rank=0, level="uppereq") == [
            frozenset({1}),
            frozenset({2}),
            frozenset({3}),
            frozenset({4}),
            frozenset({1, 2, 3, 4}),
        ]

        assert hev.skeleton(rank=2, level="uppereq") == [frozenset({1, 2, 3, 4})]

        assert hev.skeleton(rank=3, level="uppereq") == []

        assert hev.skeleton(rank=0, level="upeq") == [
            frozenset({1}),
            frozenset({2}),
            frozenset({3}),
            frozenset({4}),
            frozenset({1, 2, 3, 4}),
        ]

        assert hev.skeleton(rank=2, level="upeq") == [frozenset({1, 2, 3, 4})]

        assert hev.skeleton(rank=3, level="upeq") == []

        assert hev.skeleton(rank=0, level="downeq") == [
            frozenset({1}),
            frozenset({2}),
            frozenset({3}),
            frozenset({4}),
        ]

        assert hev.skeleton(rank=2, level="downeq") == [
            frozenset({1}),
            frozenset({2}),
            frozenset({3}),
            frozenset({4}),
            frozenset({1, 2, 3, 4}),
        ]

        assert hev.skeleton(rank=0, level="lowereq") == [
            frozenset({1}),
            frozenset({2}),
            frozenset({3}),
            frozenset({4}),
        ]

        assert hev.skeleton(rank=2, level="lowereq") == [
            frozenset({1}),
            frozenset({2}),
            frozenset({3}),
            frozenset({4}),
            frozenset({1, 2, 3, 4}),
        ]

        with pytest.raises(ValueError) as exp_exception:
            hev.skeleton(rank=2, level="should_raise_error")

        assert (
            str(exp_exception.value)
            == "level must be 'equal', 'uppereq', 'lowereq', 'upeq', 'downeq', 'uppereq', 'lower', 'up', or 'down'"
        )

    def test_hyper_edge_view_get_rank(self):
        """Test the get_rank method of the Hyperedge View Class."""
        he_1 = HyperEdge((1, 2, 3, 4), rank=2)

        CC = CombinatorialComplex([he_1], name="trial_CC")
        hev = CC._complex_set

        assert hev.get_rank([]) == 0

        assert hev.get_rank([1]) == 0

        assert hev.get_rank([1, 2, 3, 4]) == 2

        with pytest.raises(KeyError) as exp_exception:
            hev.get_rank([1, 4, 5, 8])

        assert (
            str(exp_exception.value) == "'hyperedge [1, 4, 5, 8] is not in the complex'"
        )

        he_2 = HyperEdge({}, rank=1)

        assert hev.get_rank(he_2) == 0

        he_3 = HyperEdge({1}, rank=0)

        assert hev.get_rank(he_3) == 0

        assert hev.get_rank(he_1) == 2

        he_4 = HyperEdge({8, 1, 9, 7})

        with pytest.raises(KeyError) as exp_exception:
            hev.get_rank(he_4)

        assert (
            str(exp_exception.value)
            == f"'hyperedge {he_4.elements} is not in the complex'"
        )

        assert hev.get_rank(1) == 0
        assert hev.get_rank(2) == 0

        with pytest.raises(KeyError) as exp_exception:
            hev.get_rank(6)

        assert (
            str(exp_exception.value)
            == "'hyperedge frozenset({6}) is not in the complex'"
        )

        he_5 = HyperEdge(("1", "2", "3", "4"), rank=2)

        CC_2 = CombinatorialComplex([he_5], name="trial_CC")
        hev_2 = CC_2._complex_set

        with pytest.raises(KeyError) as exp_exception:
            hev_2.get_rank("6")

        assert (
            str(exp_exception.value)
            == "\"hyperedge frozenset({'6'}) is not in the complex\""
        )

        assert hev_2.get_rank("1") == 0
        assert hev_2.get_rank("2") == 0

    def test_hyper_edge_view_get_lower_rank(self):
        """Test the _get_lower_rank method of the Hyperedge View Class."""
        CC = CombinatorialComplex([], name="trial_CC")
        hev = CC._complex_set

        assert hev._get_lower_rank(rank=1) == -1

        he_1 = HyperEdge((1, 2, 3, 4), rank=2)
        he_2 = HyperEdge((1, 2, 3, 4, 6, 7), rank=3)

        CC = CombinatorialComplex([he_1, he_2], name="trial_CC")
        hev = CC._complex_set

        assert hev._get_lower_rank(rank=3) == -1

        assert hev._get_lower_rank(rank=0) == -1

        assert hev._get_lower_rank(rank=5) == -1

        assert hev._get_lower_rank(rank=2) == 0

        with pytest.raises(ValueError, match="not in list"):
            hev._get_lower_rank(rank=1)

    def test_hyper_edge_view_get_higher_rank(self):
        """Test the _get_higher_rank method of the Hyperedge View Class."""
        CC = CombinatorialComplex([], name="trial_CC")
        hev = CC._complex_set

        assert hev._get_higher_rank(rank=1) == -1

        he_1 = HyperEdge((1, 2, 3, 4), rank=2)
        he_2 = HyperEdge((1, 2, 3, 4, 6, 7), rank=3)

        CC = CombinatorialComplex([he_1, he_2], name="trial_CC")
        hev = CC._complex_set

        assert hev._get_higher_rank(rank=3) == -1

        assert hev._get_higher_rank(rank=0) == -1

        assert hev._get_higher_rank(rank=6) == -1

        assert hev._get_higher_rank(rank=2) == 3

        with pytest.raises(ValueError, match="not in list"):
            hev._get_higher_rank(rank=1)


class TestReportViews_ColoredHyperEdgeView:
    """Test the ColoredHyperEdgeView class of the ReportViews module."""

    CHG = ColoredHyperGraph([(1, 2), (2, 3), (3, 4)], colors=["red", "green", "blue"])
    colorhg_view = CHG._complex_set

    CHG1 = ColoredHyperGraph()
    colorhg_view1 = CHG1._complex_set

    def test_getitem(self):
        """Test the getitem method of the ColoredHyperEdgeView class."""
        assert self.colorhg_view[((1, 2), 0)] == {"weight": 1}

    def test_repr(self):
        """Test the repr method."""
        assert (
            self.colorhg_view.__repr__()
            == "ColoredHyperEdgeView([((1, 2), 0), ((2, 3), 0), ((3, 4), 0)])"
        )

    def test_str(self):
        """Test the str method."""
        assert (
            str(self.colorhg_view)
            == "ColoredHyperEdgeView([((1, 2), 0), ((2, 3), 0), ((3, 4), 0)])"
        )

    def test_contains(self):
        """Test the contains method."""
        assert ((1, 2), 0) in self.colorhg_view
        assert [] not in self.colorhg_view
        assert ((2, 3), 0) in self.colorhg_view
        assert 1 in self.colorhg_view
        assert ([], 1) not in self.colorhg_view
        assert ((5, 6), 0) not in self.colorhg_view
        assert HyperEdge([1, 2]) in self.colorhg_view
        assert (HyperEdge([1, 2, 3]), 0) not in self.colorhg_view
        assert (HyperEdge([1, 2]), 0) in self.colorhg_view
        assert (HyperEdge([]), 0) not in self.colorhg_view

        # test for empty CHG
        assert HyperEdge([1, 2]) not in self.colorhg_view1
        assert (HyperEdge([1, 2]), 0) not in self.colorhg_view1

    def test_skeleton(self):
        """Test the skeleton method of ColorHyperGraphView."""
        assert self.colorhg_view.skeleton(rank=2) == []
        assert self.colorhg_view.skeleton(rank=1) == [
            (frozenset({1, 2}), 0),
            (frozenset({2, 3}), 0),
            (frozenset({3, 4}), 0),
        ]
        assert self.colorhg_view.skeleton(rank=1, store_hyperedge_key=False) == [
            frozenset({1, 2}),
            frozenset({2, 3}),
            frozenset({3, 4}),
        ]

    def test_get_rank(self):
        """Test get_rank method."""
        assert self.colorhg_view.get_rank(HyperEdge([])) == 0
        assert self.colorhg_view.get_rank(HyperEdge([1, 2])) == 1
        with pytest.raises(KeyError):
            self.colorhg_view.get_rank(HyperEdge([5, 6]))

        assert self.colorhg_view.get_rank([]) == 0
        assert self.colorhg_view.get_rank([1, 2]) == 1
        with pytest.raises(KeyError):
            self.colorhg_view.get_rank([5, 6])

        assert self.colorhg_view.get_rank(1) == 0
        with pytest.raises(KeyError):
            self.colorhg_view.get_rank(5)


class TestReportViews_SimplexView:
    """Test the SimplexView class of the ReportViews module."""

    SC = SimplicialComplex([(1, 2), (2, 3), (3, 4)])
    simplex_view = SC.simplices

    def test_getitem(self):
        """Test __getitem__ method of the SimplexView."""
        assert self.simplex_view[(1, 2)] == {}
        assert self.simplex_view[1] == {}
        with pytest.raises(KeyError):
            _ = self.simplex_view[(5,)]

    def test_str(self):
        """Test __str__ method of the SimplexView."""
        assert (
            self.simplex_view.__str__()
            == "SimplexView([(1,), (2,), (3,), (4,), (1, 2), (2, 3), (3, 4)])"
        )

    def test_shape(self) -> None:
        """Test the shape method of the SimplexView."""
        assert self.simplex_view.shape == (4, 3)


class TestReportViews_NodeView:
    """Test the NodeView class of the ReportViews module."""

    SC = SimplicialComplex([(1, 2), (2, 3), (3, 4)])
    node_view = SC.nodes
    CHG = ColoredHyperGraph([(1, 2), (2, 3), (3, 4)], colors=["red", "green", "blue"])
    node_view_1 = CHG.nodes

    def test_init_cell_type(self):
        """Test NodeView init with None cell_type."""
        with pytest.raises(ValueError):
            NodeView([(1, 2), (2, 3), (3, 4)], cell_type=None)

    def test_repr(self):
        """Test the repr of NodeView."""
        assert self.node_view.__repr__() == "NodeView([(1,), (2,), (3,), (4,)])"

    def test_getitem(self):
        """Test __getitem__ method of the NodeView."""
        assert self.node_view[Simplex([1])] == {}
        with pytest.raises(KeyError):
            _ = self.node_view[(1, 2)]

        # test for nodes of ColoredHyperGraph.
        assert self.node_view_1[(1,)] == {"weight": 1}


class TestReportViews_PathView:
    """Test the PathView class of the ReportViews module."""

    path_1 = Path((1,), name="path_1", weight=5)
    path_2 = Path((1, 2), weight=10)
    path_3 = Path((1, 2, 3))

    pc = PathComplex([path_1, path_2, path_3])
    path_view = pc.paths

    def test_get_item(self):
        """Test the __getitem__ method of the PathView class."""
        assert self.path_view.__getitem__((1,)) == {"name": "path_1", "weight": 5}
        assert self.path_view.__getitem__(1) == {"name": "path_1", "weight": 5}
        assert self.path_view.__getitem__(Path(1)) == {"name": "path_1", "weight": 5}
        assert self.path_view.__getitem__((2, 3)) == {}
        assert self.path_view.__getitem__((1, 2)) == {"weight": 10}
        assert self.path_view.__getitem__((1, 2, 3)) == {}

        with pytest.raises(KeyError):
            self.path_view.__getitem__(4)
        with pytest.raises(KeyError):
            self.path_view.__getitem__((4, 3))
        with pytest.raises(KeyError):
            self.path_view.__getitem__(Path((1, 3)))

    def test_contains(self):
        """Test the __contains__ method of the PathView class."""
        assert 1 in self.path_view
        assert (1,) in self.path_view
        assert Path(1) in self.path_view
        assert (1, 2, 3, 4) not in self.path_view
        assert Path((1, 2, 4)) not in self.path_view
        assert {(1, 2, 3)} not in self.path_view
        assert [] not in self.path_view
        assert Path([1, 2, 3, 4]) not in self.path_view

    def test_repr(self):
        """Test __repr__ method of the PathView class."""
        assert (
            self.path_view.__repr__()
            == "PathView([(1,), (2,), (3,), (1, 2), (2, 3), (1, 2, 3)])"
        )

    def test_str(self):
        """Test the __str__ method of the PathView class."""
        assert (
            self.path_view.__str__()
            == "PathView([(1,), (2,), (3,), (1, 2), (2, 3), (1, 2, 3)])"
        )



"""Abstract class for Complex and Atom."""

import abc
from collections.abc import Collection, Hashable, Iterator
from typing import Any, Self

__all__ = ["Atom", "Complex"]


class Atom[AtomCollectionType: Collection[Hashable]](abc.ABC):
    """Abstract class representing an atom in a complex.

    Parameters
    ----------
    elements : Collection[Hashable]
        The elements in the atom.
    **kwargs : keyword arguments, optional
        Additional attributes to be associated with the atom.
    """

    elements: AtomCollectionType
    name: str

    def __init__(self, elements: AtomCollectionType, **kwargs) -> None:
        self.elements = elements

        self._attributes = {}
        self._attributes.update(kwargs)

    def __eq__(self, other: Any) -> bool:
        """Return True if the given atom is equal to this atom.

        Atoms are considered equal if they have the same elements but may have
        different attributes.

        Parameters
        ----------
        other : Any
            The atom to compare.

        Returns
        -------
        bool
            Returns `True` if the given atom is equal to this atom and `False` otherwise.

        Examples
        --------
        >>> s1 = tnx.Simplex((1, 2), weight=1)
        >>> s2 = tnx.Simplex((1, 2), weight=2)
        >>> s3 = tnx.Simplex((1, 2, 3))
        >>> s1 == s2
        True
        >>> s1 == s3
        False
        """
        return type(self) is type(other) and self.elements == other.elements

    def __hash__(self) -> int:
        """Return the hash of the atom.

        Returns
        -------
        int
            The hash of the atom.
        """
        return hash(self.elements)

    def __len__(self) -> int:
        """Return the number of elements in the atom.

        Returns
        -------
        int
            The number of elements in the atom.
        """
        return len(self.elements)

    def __iter__(self):
        """Return an iterator over the elements in the atom.

        Returns
        -------
        Iterator
            Iterator over the elements in the atom.
        """
        return iter(self.elements)

    def __contains__(self, item: Any) -> bool:
        """Return True if the given element is contained in this atom.

        Parameters
        ----------
        item : Any
            The item to be checked.

        Returns
        -------
        bool
            Returns `True` if the item is contained in the element
            and `False` otherwise.
        """
        return item in self.elements

    def __getitem__(self, item: Any) -> Any:
        """Return the attribute with the given name.

        Parameters
        ----------
        item : Any
            The name of the attribute.

        Returns
        -------
        Any
            The value of the attribute.

        Raises
        ------
        KeyError
            If the attribute does not exist on this atom.
        """
        return self._attributes[item]

    def __setitem__(self, key: Any, value: Any) -> None:
        """Set the attribute with the given name to the given value.

        Parameters
        ----------
        key : Any
            The name of the attribute.
        value : Any
            The value of the attribute.
        """
        self._attributes[key] = value

    def update(self, attributes: dict) -> None:
        """Update the attributes of the atom.

        Parameters
        ----------
        attributes : dict
            The attributes to be updated.
        """
        self._attributes.update(attributes)


class Complex(abc.ABC):
    """Abstract class representing a complex.

    A complex is a space that is constructed by attaching lower-dimensional
    cells to a topological space to form a new space. The cells are attached to the space in a specific way,
    and the resulting space has a well-defined structure.

    Example of complexes:

    (1) Cell Complexes : Cell complexes can be used to represent various mathematical objects, such as graphs,
    manifolds, and discrete geometric shapes. They are useful in many areas of mathematics,
    such as algebraic topology and geometry, where they can be used to study the structure and
    properties of these objects.

    (2) Simplicial Complexes : Simplicial complexes are mathematical structures used to
    study the topological properties of shapes and spaces. They are made up of a set
    of points called vertices, and a collection of simplices (triangles, tetrahedra, etc.)
    that are connected to each other in a specific way. Each simplex is formed by a subset
    of the vertices and is considered a building block of the complex. The properties of
    the complex are determined by the combinatorial arrangement of the simplices and their
    connectivity. Simplicial complexes are used in many areas of mathematics and computer
    science, such as geometric modeling, data analysis, and machine learning.

    (3) The CombinatorialComplex class represents a combinatorial complex, which is a mathematical
    structure consisting of a set of points, a subset of the power set of points, and a ranking function
    that assigns a rank to each subset based on its size. These classes are used in many areas of mathematics
    and computer science, such as geometric modeling, data analysis, and machine learning.

    Parameters
    ----------
    **kwargs : keyword arguments, optional
        Attributes to add to the complex as key=value pairs.

    Attributes
    ----------
    complex : dict
        A dictionary that can be used to store additional information about the complex.
    """

    complex: dict[Any, Any]

    def __init__(self, **kwargs) -> None:
        """Initialize a new instance of the Complex class.

        Parameters
        ----------
        **kwargs : keyword arguments, optional
            Attributes to add to the complex as key=value pairs.
        """
        self.complex = {}
        self.complex.update(kwargs)

    @property
    @abc.abstractmethod
    def nodes(self) -> Iterator[Hashable]:
        """Return an iterator over the nodes in the complex.

        Returns
        -------
        Iterator[Hashable]
            Iterator over the nodes in the complex.
        """

    @property
    @abc.abstractmethod
    def dim(self) -> int:
        """Return dimension of the complex.

        Returns
        -------
        int
            Returns the dimension of the complex.
        """

    @property
    @abc.abstractmethod
    def shape(self) -> tuple[int, ...]:
        """Return number of cells for each rank.

        Returns
        -------
        tuple of ints
            The number of elements for each rank. If the complex is empty, an empty
            tuple is returned.
        """

    @abc.abstractmethod
    def skeleton(self, rank: int):
        """Return the atoms of given `rank` in this complex.

        Parameters
        ----------
        rank : int
            The rank of the skeleton.
        """

    @abc.abstractmethod
    def __str__(self) -> str:
        """Print basic string representation.

        Returns
        -------
        str
            Returns the string representation of the complex.
        """

    @abc.abstractmethod
    def __repr__(self) -> str:
        """Printable representation of the complex.

        Makes an attempt to return a string that would produce an object with the same value when passed to ``eval()``,
        but may not be possible for all objects.

        Returns
        -------
        str
            Returns the __repr__ representation of the complex.
        """

    @abc.abstractmethod
    def __len__(self) -> int:
        """Return number of nodes."""

    @abc.abstractmethod
    def clone(self) -> Self:
        """Clone complex."""

    @abc.abstractmethod
    def __iter__(self) -> Iterator:
        """Return an iterator over the atoms in this complex.

        Returns
        -------
        Iterator[Atom]
            Iterator over the atoms in the complex.
        """

    @abc.abstractmethod
    def __contains__(self, atom: Any) -> bool:
        """Check whether this complex contains the given atom.

        Parameters
        ----------
        atom : Any
            The atom to be checked.

        Returns
        -------
        bool
            Returns `True` if the complex contains the atom, else `False`.
        """

    @abc.abstractmethod
    def __getitem__(self, atom: Any) -> dict[Hashable, Any]:
        """Return the user-defined attributes associated with the given atom.

        Writing to the returned dictionary will update the user-defined attributes
        associated with the atom.

        Parameters
        ----------
        atom : Any
            The atom for which to return the associated user-defined attributes.

        Returns
        -------
        dict[Hashable, Any]
            The user-defined attributes associated with the given atom.

        Raises
        ------
        KeyError
            If the atom does not exist in this complex.
        """

    @abc.abstractmethod
    def remove_nodes(self, node_set: Iterator[Hashable]) -> None:
        """Remove the given nodes from the complex.

        Any elements that become invalid due to the removal of nodes are also removed.

        Parameters
        ----------
        node_set : Iterator[Hashable]
            The nodes to be removed.
        """

    @abc.abstractmethod
    def add_node(self, node: Hashable) -> None:
        """Add node to the complex.

        Parameters
        ----------
        node : Hashable
            The node to be added.
        """

    @abc.abstractmethod
    def incidence_matrix(
        self,
        rank: int,
        signed: bool = True,
        weight: str | None = None,
        index: bool = False,
    ):
        """Return incidence matrix of the complex.

        Parameters
        ----------
        rank : int
            The rank of the atoms to consider.
        signed : bool, default=True
            If True, the incidence matrix is signed, otherwise it is unsigned.
        weight : str, optional
            The name of the attribute to use as weights for the incidence matrix.
        index : bool, default=False
            If True, the incidence matrix is indexed by the nodes of the complex.
        """

    @abc.abstractmethod
    def adjacency_matrix(
        self,
        rank: int,
        signed: bool = True,
        weight: str | None = None,
        index: bool = False,
    ):
        """Return adjacency matrix of the complex.

        Parameters
        ----------
        rank : int
            The rank of the atoms to consider.
        signed : bool, default=True
            If True, the adjacency matrix is signed, otherwise it is unsigned.
        weight : str, optional
            The name of the attribute to use as weights for the adjacency matrix.
        index : bool, default=False
            If True, the adjacency matrix is indexed by the atoms of the complex.
        """

    @abc.abstractmethod
    def coadjacency_matrix(
        self,
        rank: int,
        signed: bool = True,
        weight: str | None = None,
        index: bool = False,
    ):
        """Return coadjacency matrix of the complex.

        Parameters
        ----------
        rank : int
            The rank of the atoms to consider.
        signed : bool, default=True
            If True, the adjacency matrix is signed, otherwise it is unsigned.
        weight : str, optional
            The name of the attribute to use as weights for the adjacency matrix.
        index : bool, default=False
            If True, the adjacency matrix is indexed by the atoms of the complex.
        """



"""Test graph dataset."""

import pytest

from toponetx.datasets.graph import coauthorship, karate_club


class TestGraph:
    """Test datasets utils."""

    def test_karate_club(self):
        """Test karate_club."""
        simplicial_karate_club_data = karate_club("simplicial")

        assert len(simplicial_karate_club_data.get_simplex_attributes("node_feat")) != 0
        assert len(simplicial_karate_club_data.get_simplex_attributes("edge_feat")) != 0
        assert len(simplicial_karate_club_data.get_simplex_attributes("face_feat")) != 0
        assert (
            len(simplicial_karate_club_data.get_simplex_attributes("tetrahedron_feat"))
            != 0
        )

        cell_karate_club_data = karate_club("cell")

        assert len(cell_karate_club_data.get_cell_attributes("node_feat", rank=0)) != 0
        assert len(cell_karate_club_data.get_cell_attributes("edge_feat", rank=1)) != 0
        assert len(cell_karate_club_data.get_cell_attributes("cell_feat", rank=2)) != 0

    def test_coauthorship(self):
        """Test coauthorship."""
        with pytest.warns(DeprecationWarning):
            simplicial_coauthorship_data = coauthorship()

        assert (
            len(simplicial_coauthorship_data.get_simplex_attributes("citations")) != 0
        )



"""Tests for AHORN dataset loader."""
# pyright: reportUnknownMemberType=false, reportUnknownVariableType=false, reportUnknownArgumentType=false

import networkx as nx
import pytest

from toponetx.datasets.ahorn import load_ahorn_dataset, read_ahorn_dataset

try:
    import ahorn_loader
except ImportError:
    ahorn_loader = None


@pytest.mark.skipif(
    ahorn_loader is None, reason="Optional dependency `ahorn-loader` not installed."
)
def test_load_karate_club_matches_networkx_sc() -> None:
    """Load AHORN 'karate-club' and compare with networkx.karate_club_graph."""
    SC = load_ahorn_dataset("karate-club")

    # Build graph from simplicial complex
    G_sc = SC.graph_skeleton()
    nodes_sc = set(G_sc.nodes())
    edges_sc = {tuple(sorted(e)) for e in G_sc.edges()}

    # Reference networkx graph
    G_nx = nx.karate_club_graph()
    nodes_nx = set(G_nx.nodes())
    edges_nx = {tuple(sorted(e)) for e in G_nx.edges()}

    # Basic structure equality
    assert nodes_sc == nodes_nx
    assert edges_sc == edges_nx


@pytest.mark.skipif(
    ahorn_loader is not None, reason="Optional dependency `ahorn-loader` installed."
)
def test_load_missing_dependency_raises() -> None:
    """When `ahorn-loader` is not installed, load_ahorn_dataset raises RuntimeError."""
    with pytest.raises(RuntimeError, match=r"optional `ahorn-loader`"):
        load_ahorn_dataset("karate-club")


@pytest.mark.skipif(
    ahorn_loader is not None, reason="Optional dependency `ahorn-loader` installed."
)
def test_read_missing_dependency_raises() -> None:
    """When `ahorn-loader` is not installed, read_ahorn_dataset raises RuntimeError."""
    with pytest.raises(RuntimeError, match=r"optional `ahorn-loader`"):
        read_ahorn_dataset("dummy.json")



"""
Test exterior calculus operators.

This module contains unit tests for the user-facing
`ExteriorCalculusOperators` class, verifying:

- correct construction of DEC operators (coboundaries, Hodge stars, codifferentials),
- valid shapes, basic algebraic identities, and symmetry where expected,
- error handling for invalid inputs and unsupported configurations,
- triangle-mesh metric presets route to the geometry backend and expose FEM helpers.

These tests are intentionally lightweight and avoid asserting PDE convergence.
They aim to validate correctness of API behavior and fundamental operator properties.
"""

import numpy as np
import pytest
from scipy.sparse import csr_matrix, spmatrix

from toponetx.algorithms.exterior_calculus import ExteriorCalculusOperators
from toponetx.algorithms.exterior_calculus.metric import (
    DiagonalHodgeStar,
    MetricSpec,
    TriangleMesh3DBackend,
)
from toponetx.classes.simplicial_complex import SimplicialComplex


def build_single_triangle_sc() -> SimplicialComplex:
    """Build a single-triangle simplicial complex embedded in R^3.

    Returns
    -------
    SimplicialComplex
        A simplicial complex with one triangle and vertex positions.
    """
    sc = SimplicialComplex([[0, 1, 2]])
    pos = {
        0: [0.0, 0.0, 0.0],
        1: [1.0, 0.0, 0.0],
        2: [0.0, 1.0, 0.0],
    }
    sc.set_simplex_attributes(pos, name="position")
    return sc


def build_single_triangle_sc_no_positions() -> SimplicialComplex:
    """Build a single-triangle simplicial complex without vertex positions.

    Returns
    -------
    SimplicialComplex
        A simplicial complex with one triangle and no `position` attribute.
    """
    return SimplicialComplex([[0, 1, 2]])


def build_two_triangle_square_sc() -> SimplicialComplex:
    """Build a two-triangle square patch embedded in R^3.

    This is the smallest nontrivial mesh with an interior edge shared by two
    triangles. It is useful to exercise triangle-mesh backends where some
    circumcentric dual quantities differ from a single-triangle boundary mesh.

    Returns
    -------
    SimplicialComplex
        A simplicial complex with two triangles and vertex positions.
    """
    sc = SimplicialComplex([[0, 1, 2], [0, 2, 3]])
    pos = {
        0: [0.0, 0.0, 0.0],
        1: [1.0, 0.0, 0.0],
        2: [1.0, 1.0, 0.0],
        3: [0.0, 1.0, 0.0],
    }
    sc.set_simplex_attributes(pos, name="position")
    return sc


class TestExteriorCalculusOperators:
    """Test the ExteriorCalculusOperators interface."""

    def test_dim_property(self):
        """Return correct topological dimension."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc)
        assert ops.dim == 2

    def test_d_matrix_shapes(self):
        """Return coboundary matrices with consistent shapes."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc)

        d0 = ops.d_matrix(0)
        d1 = ops.d_matrix(1)

        assert isinstance(d0, spmatrix)
        assert isinstance(d1, spmatrix)

        n0, n1, n2 = sc.shape
        assert d0.shape == (n1, n0)  # C^0 -> C^1
        assert d1.shape == (n2, n1)  # C^1 -> C^2
        assert n2 == 1

    def test_d_matrix_unsigned_is_abs_signed(self):
        """Return unsigned coboundary equal to abs(signed coboundary)."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc)

        d0_signed = ops.d_matrix(0, signed=True)
        d0_unsigned = ops.d_matrix(0, signed=False)

        assert isinstance(d0_signed, spmatrix)
        assert isinstance(d0_unsigned, spmatrix)

        # For TopoNetX, unsigned incidence should correspond to absolute values.
        assert (abs(d0_signed) != d0_unsigned).nnz == 0

    def test_d0_d1_composition_is_zero(self):
        """Satisfy the cochain complex identity d1 @ d0 = 0."""
        sc = build_two_triangle_square_sc()
        ops = ExteriorCalculusOperators(sc)

        d0 = ops.d_matrix(0, signed=True)
        d1 = ops.d_matrix(1, signed=True)

        comp = (d1 @ d0).tocsr()
        assert comp.nnz == 0

    def test_identity_metric_hodge_star(self):
        """Return identity Hodge stars when metric is identity."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc, metric="identity")

        S0 = ops.hodge_star(0)
        S1 = ops.hodge_star(1)
        S2 = ops.hodge_star(2)

        n0, n1, n2 = sc.shape
        assert S0.shape == (n0, n0)
        assert S1.shape == (n1, n1)
        assert S2.shape == (n2, n2)

        assert np.allclose(S0.diagonal(), 1.0)
        assert np.allclose(S1.diagonal(), 1.0)
        assert np.allclose(S2.diagonal(), 1.0)

    def test_hodge_star_inverse_matches_identity_metric(self):
        """Return inverse star equal to star for identity preset."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc, metric="identity")

        S0 = ops.hodge_star(0, inverse=False)
        S0_inv = ops.hodge_star(0, inverse=True)

        assert np.allclose(S0.diagonal(), S0_inv.diagonal())

    def test_codifferential_shape_and_type(self):
        """Return codifferentials with correct shapes."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc, metric="identity")

        n0, n1, n2 = sc.shape

        delta1 = ops.codifferential_matrix(1)
        delta2 = ops.codifferential_matrix(2)

        assert isinstance(delta1, spmatrix)
        assert isinstance(delta2, spmatrix)

        assert delta1.shape == (n0, n1)  # C^1 -> C^0
        assert delta2.shape == (n1, n2)  # C^2 -> C^1

    def test_codifferential_invalid_k_raises(self):
        """Raise ValueError for codifferential with k < 1."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc, metric="identity")

        with pytest.raises(ValueError):
            ops.codifferential_matrix(0)

        with pytest.raises(ValueError):
            ops.codifferential_matrix(-1)

    def test_dec_laplacian_shapes(self):
        """Return Laplacians with correct shapes for k=0,1,2."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc, metric="identity")

        n0, n1, n2 = sc.shape

        L0 = ops.dec_hodge_laplacian_matrix(0)
        L1 = ops.dec_hodge_laplacian_matrix(1)
        L2 = ops.dec_hodge_laplacian_matrix(2)

        assert isinstance(L0, spmatrix)
        assert isinstance(L1, spmatrix)
        assert isinstance(L2, spmatrix)

        assert L0.shape == (n0, n0)
        assert L1.shape == (n1, n1)
        assert L2.shape == (n2, n2)

    def test_dec_laplacian_symmetry_identity_metric(self):
        """Return symmetric DEC Laplacians for identity metric."""
        sc = build_two_triangle_square_sc()
        ops = ExteriorCalculusOperators(sc, metric="identity")

        for k in (0, 1, 2):
            Lk = ops.dec_hodge_laplacian_matrix(k)
            diff = (Lk - Lk.T).tocsr()
            assert diff.nnz == 0

    def test_dec_laplacian_psd_energy_identity_metric(self):
        """Produce nonnegative quadratic energy u^T Delta u for identity metric."""
        sc = build_two_triangle_square_sc()
        ops = ExteriorCalculusOperators(sc, metric="identity")

        L0 = ops.dec_hodge_laplacian_matrix(0).tocsr()
        n0 = L0.shape[0]
        u = np.arange(1, n0 + 1, dtype=float)

        energy = float(u @ (L0 @ u))
        assert energy >= -1e-12

    def test_invalid_k_raises(self):
        """Raise ValueError for invalid cochain degree."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc)

        with pytest.raises(ValueError):
            ops.d_matrix(-1)

        with pytest.raises(ValueError):
            ops.d_matrix(3)

        with pytest.raises(ValueError):
            ops.dec_hodge_laplacian_matrix(-1)

        with pytest.raises(ValueError):
            ops.dec_hodge_laplacian_matrix(5)

    def test_triangle_mesh_backend_basic_star(self):
        """Return valid circumcentric Hodge stars on a minimal triangle mesh."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc, metric="circumcentric", pos_name="position")

        S0 = ops.hodge_star(0)
        S1 = ops.hodge_star(1)
        S2 = ops.hodge_star(2)

        n0, n1, n2 = sc.shape
        assert S0.shape == (n0, n0)
        assert S1.shape == (n1, n1)
        assert S2.shape == (n2, n2)

        # Circumcentric dual measures are nonnegative; triangle-area reciprocal is positive.
        assert np.all(S0.diagonal() >= 0.0)
        assert np.all(S1.diagonal() >= 0.0)
        assert np.all(S2.diagonal() > 0.0)

    def test_triangle_mesh_metric_requires_positions_raises(self):
        """Raise when triangle-mesh metric is used without vertex positions."""
        sc = build_single_triangle_sc_no_positions()

        # Constructing the backend occurs during ops init; missing position labels should fail.
        with pytest.raises(KeyError):
            _ = ExteriorCalculusOperators(
                sc, metric="circumcentric", pos_name="position"
            )

    def test_triangle_mesh_backend_exposes_fem_helpers(self):
        """Expose FEM helpers only for triangle-mesh presets."""
        sc = build_two_triangle_square_sc()
        ops = ExteriorCalculusOperators(sc, metric="circumcentric", pos_name="position")

        M0 = ops.fem_mass_matrix_0()
        K0 = ops.cotan_stiffness_0()

        assert isinstance(M0, spmatrix)
        assert isinstance(K0, spmatrix)
        assert M0.shape == (4, 4)
        assert K0.shape == (4, 4)

        # Mass and stiffness should be symmetric.
        assert (M0 - M0.T).tocsr().nnz == 0
        assert (K0 - K0.T).tocsr().nnz == 0

    def test_fem_helpers_raise_without_triangle_mesh_backend(self):
        """Raise RuntimeError for FEM helpers when backend is not triangle-mesh."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc, metric="identity", pos_name="position")

        with pytest.raises(RuntimeError):
            _ = ops.fem_mass_matrix_0()

        with pytest.raises(RuntimeError):
            _ = ops.cotan_stiffness_0()

        with pytest.raises(RuntimeError):
            _ = ops.riemannian_stiffness_0()

        with pytest.raises(RuntimeError):
            _ = ops.cotan_laplacian_0()

        with pytest.raises(RuntimeError):
            _ = ops.riemannian_laplacian_0()

    def test_cotan_laplacian_lumped_and_unlumped_shapes(self):
        """Return cotan Laplacian with correct shapes for lumped/unlumped modes."""
        sc = build_two_triangle_square_sc()
        ops = ExteriorCalculusOperators(sc, metric="circumcentric", pos_name="position")

        L_lumped = ops.cotan_laplacian_0(lumped=True)
        L_unlumped = ops.cotan_laplacian_0(lumped=False)

        assert isinstance(L_lumped, spmatrix)
        assert isinstance(L_unlumped, spmatrix)
        assert L_lumped.shape == (4, 4)
        assert L_unlumped.shape == (4, 4)

    def test_riemannian_laplacian_defaults_to_isotropic(self):
        """Return anisotropic Laplacian even when no tensors are provided."""
        sc = build_two_triangle_square_sc()
        ops = ExteriorCalculusOperators(sc, metric="circumcentric", pos_name="position")

        L = ops.riemannian_laplacian_0(lumped=True)
        assert isinstance(L, spmatrix)
        assert L.shape == (4, 4)

    def test_custom_star_backend_overrides_metric_selection(self):
        """Allow explicit star_backend to override automatic metric backend."""
        sc = build_single_triangle_sc_no_positions()

        # If a custom backend is provided, ops should not attempt to build the
        # triangle-mesh backend (and thus should not require positions).
        custom_backend = DiagonalHodgeStar(metric=MetricSpec(preset="identity"))
        ops = ExteriorCalculusOperators(
            sc,
            metric="circumcentric",
            pos_name="position",
            star_backend=custom_backend,
        )

        S0 = ops.hodge_star(0)
        assert isinstance(S0, spmatrix)
        assert np.allclose(S0.diagonal(), 1.0)

    def test_metric_none_gives_identity_stars(self):
        """Return identity stars when metric support is disabled (metric=None)."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc, metric=None)

        S0 = ops.hodge_star(0)
        assert isinstance(S0, spmatrix)
        assert np.allclose(S0.diagonal(), 1.0)

    def test_hodge_star_returns_sparse_matrix_for_all_k(self):
        """Return sparse matrices for all valid cochain degrees."""
        sc = build_two_triangle_square_sc()
        ops = ExteriorCalculusOperators(sc, metric="circumcentric", pos_name="position")

        for k in (0, 1, 2):
            Sk = ops.hodge_star(k)
            assert isinstance(Sk, spmatrix)

    def test_identity_metric_returns_csr_for_d_matrix_if_backend_is_toponetx(self):
        """Return SciPy sparse matrix type from TopoNetX incidence (not necessarily CSR)."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc, metric="identity")

        d0 = ops.d_matrix(0)
        assert isinstance(d0, spmatrix)
        d0_csr = d0.tocsr()
        assert isinstance(d0_csr, csr_matrix)

    # ------------------------------------------------------------------
    # Extra coverage to hit remaining branches in ExteriorCalculusOperators
    # ------------------------------------------------------------------

    def test_normalize_metric_accepts_metricspec(self):
        """Accept MetricSpec inputs without modification."""
        ms = MetricSpec(preset="identity")
        out = ExteriorCalculusOperators._normalize_metric(ms)
        assert isinstance(out, MetricSpec)
        assert out.preset == "identity"

    def test_normalize_metric_accepts_preset_string(self):
        """Accept preset strings and convert to MetricSpec."""
        out = ExteriorCalculusOperators._normalize_metric("diagonal")
        assert isinstance(out, MetricSpec)
        assert out.preset == "diagonal"

    def test_normalize_metric_type_error(self):
        """Raise TypeError for unsupported metric types."""
        with pytest.raises(TypeError):
            _ = ExteriorCalculusOperators._normalize_metric(123)  # type: ignore[arg-type]

    def test_select_backend_diagonal_builds_diagonal_backend(self):
        """Diagonal preset builds DiagonalHodgeStar backend when weights are provided."""
        sc = build_single_triangle_sc()

        metric = MetricSpec(
            preset="diagonal",
            diagonal_weights={0: np.ones(3)},
        )

        ops = ExteriorCalculusOperators(sc, metric=metric)

        S0 = ops.hodge_star(0)
        assert isinstance(S0, csr_matrix)
        assert np.allclose(S0.diagonal(), 1.0)

    def test_select_backend_triangle_presets_build_triangle_backend(self):
        """Route triangle-mesh presets to TriangleMesh3DBackend."""
        sc = build_single_triangle_sc()
        for preset in ("barycentric_lumped", "circumcentric", "voronoi"):
            ops = ExteriorCalculusOperators(sc, metric=preset, pos_name="position")
            assert isinstance(ops.star_backend, TriangleMesh3DBackend)

    def test_select_backend_unsupported_preset_raises(self):
        """Raise ValueError for unsupported preset name."""
        sc = build_single_triangle_sc()
        with pytest.raises(ValueError):
            _ = ExteriorCalculusOperators(sc, metric="not-a-real-preset")  # type: ignore[arg-type]

    def test_euclidean_preset_constructs_triangle_backend_and_star_raises(self):
        """Construct triangle backend for 'euclidean' and raise for unsupported star preset."""
        sc = build_single_triangle_sc()
        ops = ExteriorCalculusOperators(sc, metric="euclidean", pos_name="position")

        # Selection path coverage:
        assert isinstance(ops.star_backend, TriangleMesh3DBackend)

        # TriangleMesh3DBackend.star does not implement "euclidean" for stars.
        with pytest.raises(ValueError):
            _ = ops.hodge_star(0)



r"""
Exterior calculus operators built on TopoNetX coincidence matrices.

This module exposes :class:`~toponetx.algorithms.exterior_calculus.ExteriorCalculusOperators`,
a user-facing wrapper around:

- topological coboundaries ``d_k``,
- diagonal discrete Hodge stars ``*_k``,
- codifferentials ``δ_k``,
- "up" and "down" Laplacian pieces,
- DEC Hodge Laplacians ``Δ_k``,
- the (stacked) Hodge--Dirac operator,
- Hodge decomposition utilities,
- DEC inner products and norms.

Notes
-----
**Mathematical background.**
Let ``K`` be a finite cell/simplicial complex of dimension ``d``. For each ``k``,
let ``C^k(K)`` denote the space of real-valued *k-cochains* (functions assigning a
number to each oriented k-cell). The *coboundary* operator is

.. math::

    d_k : C^k(K) \to C^{k+1}(K),

implemented by TopoNetX as a (co)incidence matrix between k- and (k+1)-cells (with
optional signs encoding orientation).

Discrete Exterior Calculus (DEC) introduces a (typically diagonal) discrete Hodge
star

.. math::

    *_k : C^k(K) \to C^{d-k}(K),

which depends on a choice of metric/geometry and defines an inner product on
cochains:

.. math::

    \langle \alpha, \beta \rangle_k := \alpha^\top (*_k)\, \beta.

Given ``d_{k-1}`` and the Hodge stars, the DEC *codifferential* is

.. math::

    \delta_k := (*_{k-1})^{-1}\, d_{k-1}^\top\, (*_k) : C^k(K) \to C^{k-1}(K).

The DEC Hodge Laplacian splits into "up" and "down" pieces:

.. math::

    \Delta_k = \underbrace{\delta_{k+1} d_k}_{\text{up}} + \underbrace{d_{k-1}\delta_k}_{\text{down}}
    : C^k(K) \to C^k(K),

with the convention that missing terms are zero at boundary degrees.

A degree-wise Hodge--Dirac operator can be written as a stacked map

.. math::

    D_k := \begin{bmatrix}\delta_k \\ d_k\end{bmatrix} : C^k(K) \to C^{k-1}(K)\oplus C^{k+1}(K),

which is useful for exposing first-order structure and for building Laplacians.

**Implementation notes.**
- ``d_k`` is obtained from :meth:`TopoNetX Complex.coincidence_matrix` as
  ``coincidence_matrix(k+1, signed=..., index=False)``.
- ``*_k`` is produced by a *backend* chosen from the user metric specification.
  If metric support is disabled (``metric=None``), then ``*_k`` defaults to the
  identity matrix.
- For triangle meshes embedded in ``R^3`` (triangle-mesh presets), additional
  surface FEM operators on 0-forms are exposed through the chosen backend:
  consistent mass matrix and stiffness matrices, plus mass-lumped Laplacians.

Examples
--------
Basic DEC Laplacian on a simplicial complex (identity metric)::

    >>> import toponetx as tnx
    >>> from toponetx.algorithms.exterior_calculus import ExteriorCalculusOperators

    >>> sc = tnx.SimplicialComplex([[0, 1, 2]])
    >>> ops = ExteriorCalculusOperators(sc, metric="identity")

    >>> d0 = ops.d_matrix(0)  # C^0 -> C^1
    >>> S1 = ops.hodge_star(1)  # *_1
    >>> delta1 = ops.codifferential_matrix(1)
    >>> L0 = ops.dec_hodge_laplacian_matrix(0)

Triangle mesh in R^3 with circumcentric stars (requires vertex positions)::

    >>> import toponetx as tnx
    >>> from toponetx.algorithms.exterior_calculus import ExteriorCalculusOperators

    >>> sc = tnx.SimplicialComplex([[0, 1, 2], [0, 2, 3]])
    >>> sc.set_simplex_attributes(
        {
            0: [0.0, 0.0, 0.0],
            1: [1.0, 0.0, 0.0],
            2: [1.0, 1.0, 0.0],
            3: [0.0, 1.0, 0.0],
        },
        name="position",
    )

    >>> ops = ExteriorCalculusOperators(sc, metric="circumcentric", pos_name="position")
    >>> S0 = ops.hodge_star(0)
    >>> M0 = ops.fem_mass_matrix_0()
    >>> K0 = ops.cotan_stiffness_0()
    >>> Lcot = ops.cotan_laplacian_0(lumped=True)

Disable metric support (all stars become identities)::

    >>> ops = ExteriorCalculusOperators(sc, metric=None)
    >>> S2 = ops.hodge_star(2)  # identity on C^2
"""

from dataclasses import dataclass, field
from typing import Literal

import numpy as np
import scipy.sparse
import scipy.sparse.linalg
from scipy.sparse import csr_matrix, identity, vstack

from .metric import (
    DiagonalHodgeStar,
    HodgeStarBackend,
    MetricPreset,
    MetricSpec,
    TriangleMesh3DBackend,
)


@dataclass
class ExteriorCalculusOperators:
    r"""Compute metric-aware DEC operators on a TopoNetX complex.

    This class is a thin, user-facing wrapper that constructs and exposes core
    operators of Discrete Exterior Calculus (DEC) on a TopoNetX complex.

    Notes
    -----
    **What it provides.**
    - Coboundary operators ``d_k`` obtained from TopoNetX coincidence matrices.
    - Discrete Hodge stars ``*_k`` from a backend determined by a metric preset
      (or an explicitly supplied backend).
    - Codifferentials ``δ_k`` defined via Hodge stars and transposed coboundaries:

      .. math::

          \delta_k := (*_{k-1})^{-1}\, d_{k-1}^\top\, (*_k).

    - "Up" and "down" Laplacian pieces:

      .. math::

          L^{\mathrm{up}}_k := \delta_{k+1} d_k,\qquad
          L^{\mathrm{down}}_k := d_{k-1} \delta_k.

    - DEC Hodge Laplacian assembled as:

      .. math::

          \Delta_k := L^{\mathrm{up}}_k + L^{\mathrm{down}}_k.

    - Degree-wise Hodge--Dirac operator (stacked map):

      .. math::

          D_k := \begin{bmatrix}\delta_k \\ d_k\end{bmatrix}.

    - DEC inner products and induced norms on cochains:

      .. math::

          \langle a,b\rangle_k := a^\top (*_k)\, b,\qquad
          \|a\|_k := \sqrt{\langle a,a\rangle_k}.

    - A practical (approximate) Hodge decomposition routine that returns exact,
      coexact, and harmonic-residual components.

    **Metric behavior.**
    Users may pass:
    - ``metric=None`` to disable metric support (all stars become identities),
    - a preset string (``MetricPreset``),
    - a :class:`~toponetx.algorithms.exterior_calculus.metric.MetricSpec`,
    - an explicit ``star_backend`` implementing
      :class:`~toponetx.algorithms.exterior_calculus.metric.HodgeStarBackend`,
      which overrides automatic selection.

    **Triangle mesh extras.**
    For triangle-mesh presets (e.g. ``"circumcentric"``, ``"voronoi"``,
    ``"barycentric_lumped"``, ``"euclidean"``), the backend is
    :class:`~toponetx.algorithms.exterior_calculus.metric.TriangleMesh3DBackend`
    and the complex must provide 3D positions under ``pos_name``. In that case,
    this wrapper exposes surface FEM helpers on 0-forms via the backend:
    mass matrix, stiffness matrices, and (mass-lumped) Laplacians.

    Attributes
    ----------
    sc : object
        Complex providing:

        - ``dim`` (maximum dimension),
        - ``skeleton(k)`` (iterable of k-cells / simplices),
        - ``coincidence_matrix(rank, signed, index)`` for coboundaries.

        For triangle-mesh presets, it must also provide
        ``get_node_attributes(pos_name)`` returning 3D vertex coordinates.
    pos_name : str
        Node attribute name storing vertex positions in ``R^3`` for triangle-mesh
        presets (default: ``"position"``).
    metric : MetricSpec | MetricPreset | None
        Metric/star specification.

        - If ``None``, metric support is disabled and stars default to identity.
        - If a preset string, it is normalized to ``MetricSpec(preset=...)``.
        - If ``MetricSpec``, it is used as-is (with ``eps`` set from this object).
    star_backend : HodgeStarBackend | None
        Optional explicit backend overriding automatic selection.
    eps : float
        Numerical safeguard for star inverses and mass-lumping.

    Examples
    --------
    Build operators on a simplicial complex and compute a Laplacian::

        >>> import toponetx as tnx
        >>> from toponetx.algorithms.exterior_calculus import ExteriorCalculusOperators

        >>> sc = tnx.SimplicialComplex([[0, 1, 2]])
        >>> ops = ExteriorCalculusOperators(sc, metric="identity")

        >>> L0 = ops.dec_hodge_laplacian_matrix(0)

    Use a triangle-mesh preset (requires 3D vertex positions)::

        >>> import toponetx as tnx
        >>> from toponetx.algorithms.exterior_calculus import ExteriorCalculusOperators

        >>> sc = tnx.SimplicialComplex([[0, 1, 2], [0, 2, 3]])
        >>> sc.set_simplex_attributes(
            {
                0: [0.0, 0.0, 0.0],
                1: [1.0, 0.0, 0.0],
                2: [1.0, 1.0, 0.0],
                3: [0.0, 1.0, 0.0],
            },
            name="position",
        )

        >>> ops = ExteriorCalculusOperators(sc, metric="circumcentric", pos_name="position")
        >>> M0 = ops.fem_mass_matrix_0()
        >>> Lcot = ops.cotan_laplacian_0(lumped=True)

    Disable metric support::

        >>> ops = ExteriorCalculusOperators(sc, metric=None)
        >>> S1 = ops.hodge_star(1)  # identity
    """

    sc: object
    pos_name: str = "position"
    metric: MetricSpec | MetricPreset | None = field(
        default_factory=lambda: MetricSpec(preset="barycentric_lumped")
    )
    star_backend: HodgeStarBackend | None = None
    eps: float = 1e-12

    def __post_init__(self) -> None:
        """Post-initialize the operator object.

        This method:

        1. Casts ``eps`` to float.
        2. Normalizes the user-supplied ``metric`` into a
           :class:`~toponetx.algorithms.exterior_calculus.metric.MetricSpec`
           (or ``None`` to disable metric support).
        3. If no explicit ``star_backend`` is provided, selects one automatically
           based on the normalized metric preset.

        Returns
        -------
        None
            This returns None.
        """
        self.eps = float(self.eps)
        self.metric = self._normalize_metric(self.metric)
        if self.star_backend is None:
            self.star_backend = self._select_backend()

    @staticmethod
    def _normalize_metric(
        metric: MetricSpec | MetricPreset | None,
    ) -> MetricSpec | None:
        """Normalize user-provided metric input into a ``MetricSpec`` or ``None``.

        Users may pass:

        - ``None``: disables metric support (stars become identities).
        - :class:`~toponetx.algorithms.exterior_calculus.metric.MetricSpec`: used directly.
        - a preset string: converted to ``MetricSpec(preset=<string>)``.

        Parameters
        ----------
        metric : MetricSpec | MetricPreset | None
            User input metric specification.

        Returns
        -------
        MetricSpec | None
            Normalized metric specification.

        Raises
        ------
        TypeError
            If ``metric`` is not ``None``, a ``MetricSpec``, or a preset string.
        """
        if metric is None:
            return None
        if isinstance(metric, MetricSpec):
            return metric
        if isinstance(metric, str):
            return MetricSpec(preset=metric)  # type: ignore[arg-type]
        raise TypeError("metric must be None, a MetricSpec, or a preset string.")

    def _select_backend(self) -> HodgeStarBackend | None:
        """Select and construct the appropriate Hodge star backend.

        The backend determines how diagonal Hodge stars ``*_k`` are computed.

        Returns
        -------
        HodgeStarBackend | None
            The selected backend instance, or ``None`` if metric support is disabled.

        Raises
        ------
        ValueError
            If the metric preset is not supported.
        """
        if self.metric is None:
            return None

        preset = self.metric.preset

        if preset in ("identity", "diagonal"):
            ms = MetricSpec(
                preset=preset,
                diagonal_weights=self.metric.diagonal_weights,
                primal_measures=self.metric.primal_measures,
                dual_measures=self.metric.dual_measures,
                tensors=self.metric.tensors,
                fn=self.metric.fn,
                eps=self.eps,
            )
            return DiagonalHodgeStar(metric=ms)

        if preset in ("barycentric_lumped", "circumcentric", "voronoi", "euclidean"):
            ms = MetricSpec(
                preset=preset,
                diagonal_weights=self.metric.diagonal_weights,
                primal_measures=self.metric.primal_measures,
                dual_measures=self.metric.dual_measures,
                tensors=self.metric.tensors,
                fn=self.metric.fn,
                eps=self.eps,
            )
            return TriangleMesh3DBackend(sc=self.sc, metric=ms, pos_name=self.pos_name)

        raise ValueError(f"Unsupported metric preset {preset!r}.")

    @property
    def dim(self) -> int:
        """Return the maximum simplex dimension of the complex.

        Returns
        -------
        int
            Maximum simplex dimension.
        """
        return self.sc.dim

    def d_matrix(self, k: int, signed: bool = True) -> csr_matrix:
        r"""Return the coboundary matrix ``d_k : C^k -> C^{k+1}``.

        Parameters
        ----------
        k : int
            Cochain degree.
        signed : bool, default=True
            If True, respect orientations. If False, take absolute values.

        Returns
        -------
        csr_matrix
            Sparse coboundary matrix.
        """
        if k < 0 or k >= self.dim + 1:
            raise ValueError(f"k must be in [0, {self.dim}], got {k}.")
        return self.sc.coincidence_matrix(k + 1, signed=signed, index=False)

    def hodge_star(self, k: int, inverse: bool = False) -> csr_matrix:
        r"""Return a discrete Hodge star matrix ``*_k``.

        Parameters
        ----------
        k : int
            Cochain degree.
        inverse : bool, default=False
            If True, return ``(*_k)^{-1}``.

        Returns
        -------
        csr_matrix
            Sparse star matrix (diagonal in the provided backends).
        """
        n_k = len(list(self.sc.skeleton(k)))
        if self.star_backend is None:
            return identity(n_k, format="csr")
        return self.star_backend.star(self.sc, k, inverse=inverse)

    def codifferential_matrix(self, k: int, signed: bool = True) -> csr_matrix:
        r"""Return the DEC codifferential matrix ``δ_k : C^k -> C^{k-1}``.

        Parameters
        ----------
        k : int
            Cochain degree. Must satisfy ``k >= 1``.
        signed : bool, default=True
            Whether to use signed coboundaries (orientation-aware).

        Returns
        -------
        csr_matrix
            Sparse codifferential matrix.

        Raises
        ------
        ValueError
            If ``k <= 0``.
        """
        if k <= 0:
            raise ValueError("codifferential requires k >= 1.")
        star_k = self.hodge_star(k, inverse=False)
        star_km1_inv = self.hodge_star(k - 1, inverse=True)
        d_km1 = self.d_matrix(k - 1, signed=signed)
        return star_km1_inv @ (d_km1.T @ star_k)

    def laplace_up_matrix(self, k: int, signed: bool = True) -> csr_matrix:
        r"""Return the up-Laplacian matrix ``L_k^up = δ_{k+1} d_k``.

        Parameters
        ----------
        k : int
            Cochain degree.
        signed : bool, default=True
            Whether to use signed coboundaries (orientation-aware).

        Returns
        -------
        csr_matrix
            Sparse up-Laplacian matrix.

        Raises
        ------
        ValueError
            If ``k`` is outside ``[0, dim]``.
        """
        if k < 0 or k > self.dim:
            raise ValueError(f"k must be in [0, {self.dim}], got {k}.")
        if k == self.dim:
            n_k = len(list(self.sc.skeleton(k)))
            return scipy.sparse.csr_matrix((n_k, n_k))
        d_k = self.d_matrix(k, signed=signed)
        delta_kp1 = self.codifferential_matrix(k + 1, signed=signed)
        return delta_kp1 @ d_k

    def laplace_down_matrix(self, k: int, signed: bool = True) -> csr_matrix:
        r"""Return the down-Laplacian matrix ``L_k^down = d_{k-1} δ_k``.

        Parameters
        ----------
        k : int
            Cochain degree.
        signed : bool, default=True
            Whether to use signed coboundaries (orientation-aware).

        Returns
        -------
        csr_matrix
            Sparse down-Laplacian matrix.

        Raises
        ------
        ValueError
            If ``k`` is outside ``[0, dim]``.
        """
        if k < 0 or k > self.dim:
            raise ValueError(f"k must be in [0, {self.dim}], got {k}.")
        if k == 0:
            n_k = len(list(self.sc.skeleton(k)))
            return scipy.sparse.csr_matrix((n_k, n_k))
        delta_k = self.codifferential_matrix(k, signed=signed)
        d_km1 = self.d_matrix(k - 1, signed=signed)
        return d_km1 @ delta_k

    def dec_hodge_laplacian_matrix(self, k: int, signed: bool = True) -> csr_matrix:
        r"""Return the DEC Hodge Laplacian matrix ``Δ_k`` on k-cochains.

        Parameters
        ----------
        k : int
            Cochain degree.
        signed : bool, default=True
            Whether to use signed coboundaries (orientation-aware).

        Returns
        -------
        csr_matrix
            Sparse Laplacian matrix.

        Raises
        ------
        ValueError
            If ``k`` is outside ``[0, dim]``.
        """
        if k < 0 or k > self.dim:
            raise ValueError(f"k must be in [0, {self.dim}], got {k}.")
        return self.laplace_up_matrix(k, signed=signed) + self.laplace_down_matrix(
            k, signed=signed
        )

    def hodge_dirac_matrix(self, k: int, signed: bool = True) -> csr_matrix:
        r"""Return the degree-k Hodge--Dirac operator matrix (stacked).

        This returns

        .. math::

            D_k := \begin{bmatrix} \delta_k \\ d_k \end{bmatrix},

        which maps ``C^k`` into ``C^{k-1} \oplus C^{k+1}`` (with missing blocks omitted
        at boundary degrees).

        Parameters
        ----------
        k : int
            Cochain degree.
        signed : bool, default=True
            Whether to use signed coboundaries (orientation-aware).

        Returns
        -------
        csr_matrix
            Sparse matrix with shape ``(n_{k-1} + n_{k+1}, n_k)`` (or smaller at
            boundary degrees).

        Raises
        ------
        ValueError
            If ``k`` is outside ``[0, dim]``.
        """
        if k < 0 or k > self.dim:
            raise ValueError(f"k must be in [0, {self.dim}], got {k}.")

        blocks: list[csr_matrix] = []

        if k > 0:
            blocks.append(self.codifferential_matrix(k, signed=signed))
        if k < self.dim:
            blocks.append(self.d_matrix(k, signed=signed))

        if not blocks:
            n_k = len(list(self.sc.skeleton(k)))
            return scipy.sparse.csr_matrix((0, n_k))

        return vstack(blocks, format="csr")

    def inner_product(self, k: int, a: np.ndarray, b: np.ndarray) -> float:
        r"""Compute the DEC inner product ``<a, b>_k = a^T (*_k) b``.

        Parameters
        ----------
        k : int
            Cochain degree.
        a : np.ndarray
            A k-cochain (1D array of length ``n_k``).
        b : np.ndarray
            A k-cochain (1D array of length ``n_k``).

        Returns
        -------
        float
            The scalar inner product value.

        Raises
        ------
        ValueError
            If input shapes do not match ``n_k``.
        """
        a1 = np.asarray(a, dtype=float).reshape(-1)
        b1 = np.asarray(b, dtype=float).reshape(-1)
        n_k = len(list(self.sc.skeleton(k)))
        if a1.shape[0] != n_k or b1.shape[0] != n_k:
            raise ValueError(
                f"Expected a and b to have length n_k={n_k}; got {a1.shape[0]} and {b1.shape[0]}."
            )
        S = self.hodge_star(k, inverse=False)
        return float(a1 @ (S @ b1))

    def norm(self, k: int, a: np.ndarray) -> float:
        r"""Compute the DEC norm ``||a||_k = sqrt(<a, a>_k)``.

        Parameters
        ----------
        k : int
            Cochain degree.
        a : np.ndarray
            A k-cochain (1D array of length ``n_k``).

        Returns
        -------
        float
            The induced norm value.

        Raises
        ------
        ValueError
            If the input shape does not match ``n_k``.
        """
        val = self.inner_product(k, a, a)
        return float(np.sqrt(max(val, 0.0)))

    @staticmethod
    def _solve_linear_system(
        A: csr_matrix,
        b: np.ndarray,
        *,
        method: Literal["cg", "spsolve"] = "cg",
        tol: float = 1e-10,
        maxiter: int | None = None,
    ) -> np.ndarray:
        """Solve a sparse linear system ``A x = b``.

        Parameters
        ----------
        A : csr_matrix
            Sparse square system matrix.
        b : np.ndarray
            Right-hand side vector.
        method : {"cg", "spsolve"}, default="cg"
            Solver choice. Use ``"cg"`` for (approximately) SPD systems.
        tol : float, default=1e-10
            Convergence tolerance for iterative solvers.
        maxiter : int, optional
            Maximum iterations for iterative solvers.

        Returns
        -------
        np.ndarray
            Solution vector.

        Raises
        ------
        RuntimeError
            If the iterative solver fails to converge.
        """
        b1 = np.asarray(b, dtype=float).reshape(-1)
        if method == "spsolve":
            x = scipy.sparse.linalg.spsolve(A, b1)
            return np.asarray(x, dtype=float).reshape(-1)

        x, info = scipy.sparse.linalg.cg(A, b1, tol=tol, maxiter=maxiter)
        if info != 0:
            raise RuntimeError(f"cg did not converge (info={info}).")
        return np.asarray(x, dtype=float).reshape(-1)

    def hodge_decomposition(
        self,
        k: int,
        x: np.ndarray,
        *,
        signed: bool = True,
        solver: Literal["cg", "spsolve"] = "cg",
        tol: float = 1e-10,
        maxiter: int | None = None,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        r"""Compute an (approximate) Hodge decomposition of a k-cochain.

        This method computes vectors ``x_exact``, ``x_coexact``, and ``x_harmonic`` such that

        .. math::

            x \approx x_{\mathrm{exact}} + x_{\mathrm{coexact}} + x_{\mathrm{harmonic}}.

        The components are computed as *star-orthogonal* least-squares projections:

        - Exact projection onto ``im(d_{k-1})`` (if ``k>0``),
          by solving the normal equations with the ``*_k`` inner product.
        - Coexact projection onto ``im(δ_{k+1})`` (if ``k<dim``),
          similarly via the ``*_k`` inner product.
        - Harmonic remainder ``x_harmonic = x - x_exact - x_coexact``.

        Notes
        -----
        - On closed manifolds (no boundary) with compatible DEC choices, the exact and coexact
          subspaces are orthogonal in the DEC inner product, and the remainder lies close to
          the harmonic subspace ``ker(Δ_k)``.
        - On complexes with boundary or in degenerate settings, this remains a useful, robust
          computational decomposition, but the remainder should be interpreted as a residual.

        Parameters
        ----------
        k : int
            Cochain degree.
        x : np.ndarray
            Input k-cochain (1D array of length ``n_k``).
        signed : bool
            Whether to use signed coboundaries (orientation-aware).
        solver : {"cg", "spsolve"}, default="cg"
            Linear solver used for the normal equations.
        tol : float, default=1e-10
            Convergence tolerance for iterative solvers.
        maxiter : int, optional
            Maximum iterations for iterative solvers.

        Returns
        -------
        tuple[np.ndarray, np.ndarray, np.ndarray]
            Tuple ``(x_exact, x_coexact, x_harmonic)``, each a 1D array of length ``n_k``.

        Raises
        ------
        ValueError
            If ``k`` is outside ``[0, dim]`` or if ``x`` has the wrong length.
        RuntimeError
            If an iterative solver fails to converge.
        """
        if k < 0 or k > self.dim:
            raise ValueError(f"k must be in [0, {self.dim}], got {k}.")

        xk = np.asarray(x, dtype=float).reshape(-1)
        n_k = len(list(self.sc.skeleton(k)))
        if xk.shape[0] != n_k:
            raise ValueError(f"Expected x to have length n_k={n_k}, got {xk.shape[0]}.")

        S_k = self.hodge_star(k, inverse=False)

        x_exact = np.zeros_like(xk)
        if k > 0:
            d_km1 = self.d_matrix(k - 1, signed=signed)
            A = d_km1.T @ (S_k @ d_km1)
            b = d_km1.T @ (S_k @ xk)
            if A.shape[0] > 0 and A.nnz > 0:
                a = self._solve_linear_system(
                    A.tocsr(),
                    np.asarray(b).reshape(-1),
                    method=solver,
                    tol=tol,
                    maxiter=maxiter,
                )
                x_exact = d_km1 @ a

        x_coexact = np.zeros_like(xk)
        if k < self.dim:
            delta_kp1 = self.codifferential_matrix(k + 1, signed=signed)
            A = delta_kp1.T @ (S_k @ delta_kp1)
            b = delta_kp1.T @ (S_k @ xk)
            if A.shape[0] > 0 and A.nnz > 0:
                y = self._solve_linear_system(
                    A.tocsr(),
                    np.asarray(b).reshape(-1),
                    method=solver,
                    tol=tol,
                    maxiter=maxiter,
                )
                x_coexact = delta_kp1 @ y

        x_harm = xk - x_exact - x_coexact
        return x_exact, x_coexact, x_harm

    def fem_mass_matrix_0(self) -> csr_matrix:
        r"""Return the consistent P1 FEM mass matrix on 0-forms.

        Returns
        -------
        csr_matrix
            Vertex-by-vertex mass matrix.

        Raises
        ------
        RuntimeError
            If the current backend is not the triangle-mesh backend.
        """
        if not isinstance(self.star_backend, TriangleMesh3DBackend):
            raise RuntimeError(
                "fem_mass_matrix_0 requires a triangle-mesh metric preset."
            )
        return self.star_backend.fem_mass_matrix_0()

    def cotan_stiffness_0(self) -> csr_matrix:
        r"""Return the surface FEM stiffness matrix on 0-forms.

        Returns
        -------
        csr_matrix
            Vertex-by-vertex stiffness matrix.

        Raises
        ------
        RuntimeError
            If the current backend is not the triangle-mesh backend.
        """
        if not isinstance(self.star_backend, TriangleMesh3DBackend):
            raise RuntimeError(
                "cotan_stiffness_0 requires a triangle-mesh metric preset."
            )
        return self.star_backend.cotan_stiffness_0()

    def riemannian_stiffness_0(self) -> csr_matrix:
        r"""Return anisotropic surface stiffness matrix on 0-forms.

        Returns
        -------
        csr_matrix
            Vertex-by-vertex anisotropic stiffness matrix.

        Raises
        ------
        RuntimeError
            If the current backend is not the triangle-mesh backend.
        """
        if not isinstance(self.star_backend, TriangleMesh3DBackend):
            raise RuntimeError(
                "riemannian_stiffness_0 requires a triangle-mesh metric preset."
            )
        return self.star_backend.riemannian_stiffness_0()

    def cotan_laplacian_0(self, *, lumped: bool = True) -> csr_matrix:
        r"""Return a cotangent Laplacian operator on 0-forms.

        Parameters
        ----------
        lumped : bool, default=True
            If True, return ``M_lumped^{-1} K``. Otherwise return ``K``.

        Returns
        -------
        csr_matrix
            Laplacian-like operator.

        Raises
        ------
        RuntimeError
            If the current backend is not the triangle-mesh backend.
        """
        if not isinstance(self.star_backend, TriangleMesh3DBackend):
            raise RuntimeError(
                "cotan_laplacian_0 requires a triangle-mesh metric preset."
            )
        return self.star_backend.cotan_laplacian_0(lumped=lumped)

    def riemannian_laplacian_0(self, *, lumped: bool = True) -> csr_matrix:
        r"""Return an anisotropic Laplacian operator on 0-forms.

        Parameters
        ----------
        lumped : bool, default=True
            If True, return ``M_lumped^{-1} K(G)``. Otherwise return ``K(G)``.

        Returns
        -------
        csr_matrix
            Laplacian-like operator.

        Raises
        ------
        RuntimeError
            If the current backend is not the triangle-mesh backend.
        """
        if not isinstance(self.star_backend, TriangleMesh3DBackend):
            raise RuntimeError(
                "riemannian_laplacian_0 requires a triangle-mesh metric preset."
            )
        return self.star_backend.riemannian_laplacian_0(lumped=lumped)
